import numpy as np
import coolest
from lenstronomy.Util.param_util import phi_q2_ellipticity, shear_polar2cartesian


def shapelet_amp_coolest_to_lenstronomy(value):
    """Transforms shapelets coefficients from COOLEST conventions (x to the right) to
    lenstronomy conventions (x following ra, to the left)

    :param value: amplitude of the shapelet (float or np.array) in COOLEST conventions
    :return: amplitude of the shapelet (float or np.array) in lenstronomy conventions
    """
    if value is None:
        return None
    else:
        n = 0  # this is the shapelet order
        k = 0  # this is the index of a coefficient for a given order
        new_value = []
        for coeff in value:
            if n % 2 == 0:
                if k % 2 == 1:
                    coeff = -coeff
            elif n % 2 == 1:
                if k % 2 == 0:
                    coeff = -coeff
            new_value.append(coeff)
            k += 1
            if k == n + 1:
                n += 1
                k = 0
        return new_value


def degree_coolest_to_radian_lenstronomy(value):
    """
    Transform an angle in degree in COOLEST conventions (from y to negative x - aka East of North)
    into an angle in radian in lenstronomy conventions (from x to y with x pointing to the left - aka North of East)

    :param value: float, angle in COOLEST conventions
    :return: float, angle in lenstronomy conventions
    """
    if value is None:
        return None
    else:
        lenstro_oriented_degree = -value + 90.0
        if lenstro_oriented_degree >= 180.0:
            lenstro_oriented_degree -= 180.0
        elif lenstro_oriented_degree < 0.0:
            lenstro_oriented_degree += 180
        return lenstro_oriented_degree * np.pi / 180.0


def qphi_coolest_to_e1e2_lenstronomy(q, phi):
    """Transform q and phi (axis ratio, position angle East-of-North) to e1,e2 in
    lenstronomy.

    :param q: float, axis ratio
    :param phi: float, position angle in COOLEST conventions
    :return: e1, e2, lenstronomy usual ellipticity parameters
    """
    if None in [q, phi]:
        return None, None
    else:
        angle = degree_coolest_to_radian_lenstronomy(phi)
        e1, e2 = phi_q2_ellipticity(angle, q)
        return e1, e2


def gamma_phi_coolest_to_g1_g2_lenstronomy(gamma_ext, phi_ext):
    """Transform gamma_ext and phi_ext (shear strength, position angle East-of-North) to
    gamma1,gamma2 in lenstronomy.

    :param gamma_ext: float, shear strenght
    :param phi_ext: float, shear angle in COOLEST conventions
    :return: gamma1, gamma2, lenstronomy usual shear parameters
    """
    if None in [gamma_ext, phi_ext]:
        return None, None
    else:
        angle = degree_coolest_to_radian_lenstronomy(phi_ext)
        gamma1, gamma2 = shear_polar2cartesian(angle, gamma_ext)
        return gamma1, gamma2


def ellibounds_coolest_to_lenstronomy(q_down, q_up, phi_down, phi_up):
    """Transforms upper and lower bounds on coolest ellipticity parameters (q, phi)
    towards lenstronomy bound on e1, e2 The mapping can not be perfect but it's the best
    we can do.

    :param q_down: float, lower bound of axis ratio
    :param q_up: float, upper bound of axis ratio
    :param phi_down: float, lower bound of position angle in COOLEST conventions
    :param phi_up: float, upper bound of position angle in COOLEST conventions
    :return: e1_down, e1_up, e2_down, e2_up, bounds for lenstronomy usual ellipticity
        parameters
    """
    if None in [q_down, q_up, phi_down, phi_up]:
        return None, None, None, None
    else:
        e1_down = -(1 - q_down) / (1 + q_down)
        e2_down = -(1 - q_down) / (1 + q_down)

        e1_up = (1 - q_down) / (1 + q_down)
        e2_up = (1 - q_down) / (1 + q_down)

        # WE LOOSE THE ANGLE INFORMATION AND THE OTHER INFO IS NOT WELL TRANSLATED (best we can do)

        return e1_down, e1_up, e2_down, e2_up


def shearbounds_coolest_to_lenstronomy(
    gamma_ext_down, gamma_ext_up, phi_ext_down, phi_ext_up
):
    """Transforms upper and lower bounds on coolest shear parameters (gamma_ext,
    phi_ext) towards lenstronomy bounds on gamma_1, gamma_2 The mapping can not be
    perfect but it's the best we can do.

    :param gamma_ext_down: float, lower bound of shear strenght
    :param gamma_ext_up: float, upper bound of shear strenght
    :param phi_ext_down: float, lower bound of shear position angle in COOLEST
        conventions
    :param phi_ext_up: float, upper bound of shear position angle in COOLEST conventions
    :return: gamma1_down, gamma1_up, gamma2_down, gamma2_up ; bounds for lenstronomy
        usual shear parameters
    """
    if None in [gamma_ext_down, gamma_ext_up, phi_ext_down, phi_ext_up]:
        return None, None, None, None
    else:
        gamma1_down = -gamma_ext_up
        gamma2_down = -gamma_ext_up

        gamma1_up = gamma_ext_up
        gamma2_up = gamma_ext_up

        # WE LOOSE THE ANGLE INFORMATION AND THE OTHER INFO IS NOT WELL TRANSLATED (best we can do)

        return gamma1_down, gamma1_up, gamma2_down, gamma2_up


def update_kwargs_shear(
    shear_idx,
    lens_model_list,
    kwargs_lens,
    kwargs_lens_init,
    kwargs_lens_up,
    kwargs_lens_down,
    kwargs_lens_fixed,
    kwargs_lens_sigma,
    cleaning=False,
):
    """
    Update the lens model list and kwargs with SHEAR mass model (gamma_ext - phi_ext)

    :param shear_idx: coolest.template.classes.profiles.mass.ExternalShear object
    :param lens_model_list: the usual lenstronomy lens_model_list
    :param kwargs_lens: the usual lenstronomy kwargs
    :param kwargs_lens_init: the usual lenstronomy kwargs
    :param kwargs_lens_up: the usual lenstronomy kwargs
    :param kwargs_lens_down: the usual lenstronomy kwargs
    :param kwargs_lens_fixed: the usual lenstronomy kwargs
    :param kwargs_lens_sigma: the usual lenstronomy kwargs
    :param cleaning: bool, if True, will update the empty fields with default values + cleans the kwargs_fixed
    :return: updated list and kwargs
    """
    lens_model_list.append("SHEAR")
    for shear_name, shear_param in shear_idx.parameters.items():
        if shear_name == "gamma_ext":
            gammaext = getattr(shear_param.point_estimate, "value")
            gammaext_up = getattr(shear_param.definition_range, "max_value")
            gammaext_down = getattr(shear_param.definition_range, "min_value")
            gammaext_fixed = gammaext if getattr(shear_param, "fixed") else None
        elif shear_name == "phi_ext":
            psiext = getattr(shear_param.point_estimate, "value")
            psiext_up = getattr(shear_param.definition_range, "max_value")
            psiext_down = getattr(shear_param.definition_range, "min_value")
            psiext_fixed = psiext if getattr(shear_param, "fixed") else None
        else:
            print(f"{shear_name} not known")
    gamma1, gamma2 = gamma_phi_coolest_to_g1_g2_lenstronomy(gammaext, psiext)
    gamma1_fixed, gamma2_fixed = gamma_phi_coolest_to_g1_g2_lenstronomy(
        gammaext_fixed, psiext_fixed
    )
    gamma1_down, gamma1_up, gamma2_down, gamma2_up = shearbounds_coolest_to_lenstronomy(
        gammaext_down, gammaext_up, psiext_down, psiext_up
    )

    kw_1 = {"gamma1": gamma1, "gamma2": gamma2, "ra_0": 0.0, "dec_0": 0.0}
    kw_up_1 = {"gamma1": gamma1_up, "gamma2": gamma2_up, "ra_0": 0.0, "dec_0": 0.0}
    kw_down_1 = {
        "gamma1": gamma1_down,
        "gamma2": gamma2_down,
        "ra_0": 0.0,
        "dec_0": 0.0,
    }
    kw_fixed_1 = {
        "gamma1": gamma1_fixed,
        "gamma2": gamma2_fixed,
        "ra_0": 0.0,
        "dec_0": 0.0,
    }

    kw_ = kw_1.copy()
    kw_init = kw_1.copy()
    kw_up = kw_up_1.copy()
    kw_down = kw_down_1.copy()
    kw_fixed = kw_fixed_1.copy()

    if cleaning is True:
        kw_init_default = {"gamma1": 0.0, "gamma2": -0.0, "ra_0": 0.0, "dec_0": 0.0}
        kw_up_default = {"gamma1": 0.5, "gamma2": 0.5, "ra_0": 100, "dec_0": 100}
        kw_down_default = {"gamma1": -0.5, "gamma2": -0.5, "ra_0": -100, "dec_0": -100}
        for key, val in kw_1.items():
            if val is None:
                kw_init[key] = kw_init_default[key]
        for key, val in kw_up_1.items():
            if val is None:
                kw_up[key] = kw_up_default[key]
        for key, val in kw_down_1.items():
            if val is None:
                kw_down[key] = kw_down_default[key]
        for key, val in kw_fixed_1.items():
            if val is None:
                del kw_fixed[key]

    kwargs_lens.append(kw_)
    kwargs_lens_init.append(kw_init)
    kwargs_lens_up.append(kw_up)
    kwargs_lens_down.append(kw_down)
    kwargs_lens_fixed.append(kw_fixed)
    kwargs_lens_sigma.append({"gamma1": 0.1, "gamma2": 0.1, "ra_0": 0.0, "dec_0": 0.0})
    print("\t Shear correctly added")

    return


def update_kwargs_pemd(
    mass,
    lens_model_list,
    kwargs_lens,
    kwargs_lens_init,
    kwargs_lens_up,
    kwargs_lens_down,
    kwargs_lens_fixed,
    kwargs_lens_sigma,
    cleaning=False,
):
    """Update the lens list and kwargs with PEMD mass model.

    :param mass: coolest.template.classes.profiles.mass.PEMD object
    :param lens_model_list: the usual lenstronomy lens_model_list
    :param kwargs_lens: the usual lenstronomy kwargs
    :param kwargs_lens_init: the usual lenstronomy kwargs
    :param kwargs_lens_up: the usual lenstronomy kwargs
    :param kwargs_lens_down: the usual lenstronomy kwargs
    :param kwargs_lens_fixed: the usual lenstronomy kwargs
    :param kwargs_lens_sigma: the usual lenstronomy kwargs
    :param cleaning: bool, if True, will update the empty fields with default values +
        cleans the kwargs_fixed
    :return: updated list and kwargs
    """
    lens_model_list.append("PEMD")
    for mass_name, mass_param in mass.parameters.items():
        if mass_name == "theta_E":
            te = getattr(mass_param.point_estimate, "value")
            te_up = getattr(mass_param.definition_range, "max_value")
            te_down = getattr(mass_param.definition_range, "min_value")
            te_fixed = te if getattr(mass_param, "fixed") else None
        elif mass_name == "gamma":
            gamma = getattr(mass_param.point_estimate, "value")
            gamma_up = getattr(mass_param.definition_range, "max_value")
            gamma_down = getattr(mass_param.definition_range, "min_value")
            gamma_fixed = gamma if getattr(mass_param, "fixed") else None
        elif mass_name == "q":
            q = getattr(mass_param.point_estimate, "value")
            q_up = getattr(mass_param.definition_range, "max_value")
            q_down = getattr(mass_param.definition_range, "min_value")
            q_fixed = q if getattr(mass_param, "fixed") else None
        elif mass_name == "phi":
            phi = getattr(mass_param.point_estimate, "value")
            phi_up = getattr(mass_param.definition_range, "max_value")
            phi_down = getattr(mass_param.definition_range, "min_value")
            phi_fixed = phi if getattr(mass_param, "fixed") else None
        elif mass_name == "center_x":
            center_x = (
                -getattr(mass_param.point_estimate, "value")
                if getattr(mass_param.point_estimate, "value") is not None
                else None
            )
            center_x_up = getattr(mass_param.definition_range, "max_value")
            center_x_down = getattr(mass_param.definition_range, "min_value")
            center_x_fixed = center_x if getattr(mass_param, "fixed") else None
        elif mass_name == "center_y":
            center_y = getattr(mass_param.point_estimate, "value")
            center_y_up = getattr(mass_param.definition_range, "max_value")
            center_y_down = getattr(mass_param.definition_range, "min_value")
            center_y_fixed = center_y if getattr(mass_param, "fixed") else None
        else:
            print(f"{mass_name} not known")

    e1, e2 = qphi_coolest_to_e1e2_lenstronomy(q, phi)
    e1_fixed, e2_fixed = qphi_coolest_to_e1e2_lenstronomy(q_fixed, phi_fixed)
    e1_down, e1_up, e2_down, e2_up = ellibounds_coolest_to_lenstronomy(
        q_down, q_up, phi_down, phi_up
    )

    kw_1 = {
        "theta_E": te,
        "gamma": gamma,
        "e1": e1,
        "e2": e2,
        "center_x": center_x,
        "center_y": center_y,
    }
    kw_up_1 = {
        "theta_E": te_up,
        "gamma": gamma_up,
        "e1": e1_up,
        "e2": e2_up,
        "center_x": center_x_up,
        "center_y": center_y_up,
    }
    kw_down_1 = {
        "theta_E": te_down,
        "gamma": gamma_down,
        "e1": e1_down,
        "e2": e2_down,
        "center_x": center_x_down,
        "center_y": center_y_down,
    }
    kw_fixed_1 = {
        "theta_E": te_fixed,
        "gamma": gamma_fixed,
        "e1": e1_fixed,
        "e2": e2_fixed,
        "center_x": center_x_fixed,
        "center_y": center_y_fixed,
    }

    kw_ = kw_1.copy()
    kw_init = kw_1.copy()
    kw_up = kw_up_1.copy()
    kw_down = kw_down_1.copy()
    kw_fixed = kw_fixed_1.copy()

    if cleaning is True:
        kw_init_default = {
            "theta_E": 1.0,
            "gamma": 2.0,
            "e1": 0.0,
            "e2": -0.0,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        kw_up_default = {
            "theta_E": 100,
            "gamma": 2.5,
            "e1": 0.5,
            "e2": 0.5,
            "center_x": 100,
            "center_y": 100,
        }
        kw_down_default = {
            "theta_E": 0,
            "gamma": 1.5,
            "e1": -0.5,
            "e2": -0.5,
            "center_x": -100,
            "center_y": -100,
        }
        for key, val in kw_1.items():
            if val is None:
                kw_init[key] = kw_init_default[key]
        for key, val in kw_up_1.items():
            if val is None:
                kw_up[key] = kw_up_default[key]
        for key, val in kw_down_1.items():
            if val is None:
                kw_down[key] = kw_down_default[key]
        for key, val in kw_fixed_1.items():
            if val is None:
                del kw_fixed[key]

    kwargs_lens.append(kw_)
    kwargs_lens_init.append(kw_init)
    kwargs_lens_up.append(kw_up)
    kwargs_lens_down.append(kw_down)
    kwargs_lens_fixed.append(kw_fixed)
    kwargs_lens_sigma.append(
        {
            "theta_E": 1.5,
            "gamma": 0.2,
            "e1": 0.3,
            "e2": 0.3,
            "center_x": 0.5,
            "center_y": 0.5,
        }
    )

    print("\t PEMD correctly added")

    return


def update_kwargs_sie(
    mass,
    lens_model_list,
    kwargs_lens,
    kwargs_lens_init,
    kwargs_lens_up,
    kwargs_lens_down,
    kwargs_lens_fixed,
    kwargs_lens_sigma,
    cleaning=False,
):
    """Update the lens list and kwargs with SIE mass model.

    :param mass: coolest.template.classes.profiles.mass.SIE object
    :param lens_model_list: the usual lenstronomy lens_model_list
    :param kwargs_lens: the usual lenstronomy kwargs
    :param kwargs_lens_init: the usual lenstronomy kwargs
    :param kwargs_lens_up: the usual lenstronomy kwargs
    :param kwargs_lens_down: the usual lenstronomy kwargs
    :param kwargs_lens_fixed: the usual lenstronomy kwargs
    :param kwargs_lens_sigma: the usual lenstronomy kwargs
    :param cleaning: bool, if True, will update the empty fields with default values +
        cleans the kwargs_fixed
    :return: updated list and kwargs
    """
    lens_model_list.append("SIE")
    for mass_name, mass_param in mass.parameters.items():
        if mass_name == "theta_E":
            te = getattr(mass_param.point_estimate, "value")
            te_up = getattr(mass_param.definition_range, "max_value")
            te_down = getattr(mass_param.definition_range, "min_value")
            te_fixed = te if getattr(mass_param, "fixed") else None
        elif mass_name == "q":
            q = getattr(mass_param.point_estimate, "value")
            q_up = getattr(mass_param.definition_range, "max_value")
            q_down = getattr(mass_param.definition_range, "min_value")
            q_fixed = q if getattr(mass_param, "fixed") else None
        elif mass_name == "phi":
            phi = getattr(mass_param.point_estimate, "value")
            phi_up = getattr(mass_param.definition_range, "max_value")
            phi_down = getattr(mass_param.definition_range, "min_value")
            phi_fixed = phi if getattr(mass_param, "fixed") else None
        elif mass_name == "center_x":
            center_x = (
                -getattr(mass_param.point_estimate, "value")
                if getattr(mass_param.point_estimate, "value") is not None
                else None
            )
            center_x_up = getattr(mass_param.definition_range, "max_value")
            center_x_down = getattr(mass_param.definition_range, "min_value")
            center_x_fixed = center_x if getattr(mass_param, "fixed") else None
        elif mass_name == "center_y":
            center_y = getattr(mass_param.point_estimate, "value")
            center_y_up = getattr(mass_param.definition_range, "max_value")
            center_y_down = getattr(mass_param.definition_range, "min_value")
            center_y_fixed = center_y if getattr(mass_param, "fixed") else None
        else:
            print(f"{mass_name} not known")

    e1, e2 = qphi_coolest_to_e1e2_lenstronomy(q, phi)
    e1_fixed, e2_fixed = qphi_coolest_to_e1e2_lenstronomy(q_fixed, phi_fixed)
    e1_down, e1_up, e2_down, e2_up = ellibounds_coolest_to_lenstronomy(
        q_down, q_up, phi_down, phi_up
    )

    kw_1 = {
        "theta_E": te,
        "e1": e1,
        "e2": e2,
        "center_x": center_x,
        "center_y": center_y,
    }
    kw_up_1 = {
        "theta_E": te_up,
        "e1": e1_up,
        "e2": e2_up,
        "center_x": center_x_up,
        "center_y": center_y_up,
    }
    kw_down_1 = {
        "theta_E": te_down,
        "e1": e1_down,
        "e2": e2_down,
        "center_x": center_x_down,
        "center_y": center_y_down,
    }
    kw_fixed_1 = {
        "theta_E": te_fixed,
        "e1": e1_fixed,
        "e2": e2_fixed,
        "center_x": center_x_fixed,
        "center_y": center_y_fixed,
    }

    kw_ = kw_1.copy()
    kw_init = kw_1.copy()
    kw_up = kw_up_1.copy()
    kw_down = kw_down_1.copy()
    kw_fixed = kw_fixed_1.copy()

    if cleaning is True:
        kw_init_default = {
            "theta_E": 1.0,
            "e1": 0.0,
            "e2": -0.0,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        kw_up_default = {
            "theta_E": 100,
            "e1": 0.5,
            "e2": 0.5,
            "center_x": 100,
            "center_y": 100,
        }
        kw_down_default = {
            "theta_E": 0,
            "e1": -0.5,
            "e2": -0.5,
            "center_x": -100,
            "center_y": -100,
        }
        for key, val in kw_1.items():
            if val is None:
                kw_init[key] = kw_init_default[key]
        for key, val in kw_up_1.items():
            if val is None:
                kw_up[key] = kw_up_default[key]
        for key, val in kw_down_1.items():
            if val is None:
                kw_down[key] = kw_down_default[key]
        for key, val in kw_fixed_1.items():
            if val is None:
                del kw_fixed[key]

    kwargs_lens.append(kw_)
    kwargs_lens_init.append(kw_init)
    kwargs_lens_up.append(kw_up)
    kwargs_lens_down.append(kw_down)
    kwargs_lens_fixed.append(kw_fixed)
    kwargs_lens_sigma.append(
        {"theta_E": 1.5, "e1": 0.3, "e2": 0.3, "center_x": 0.5, "center_y": 0.5}
    )

    print("\t SIE correctly added")

    return


def update_kwargs_sersic(
    light,
    light_model_list,
    kwargs_light,
    kwargs_light_init,
    kwargs_light_up,
    kwargs_light_down,
    kwargs_light_fixed,
    kwargs_light_sigma,
    cleaning=False,
):
    """Update the source list and kwargs with SERSIC_ELLISPE light model.

    :param light: coolest.template.classes.profiles.light.Sersic object
    :param light_model_list: the usual lenstronomy lens_light_model_list or
        source_light_model_list
    :param kwargs_light: the usual lenstronomy kwargs
    :param kwargs_light_init: the usual lenstronomy kwargs
    :param kwargs_light_up: the usual lenstronomy kwargs
    :param kwargs_light_down: the usual lenstronomy kwargs
    :param kwargs_light_fixed: the usual lenstronomy kwargs
    :param kwargs_light_sigma: the usual lenstronomy kwargs
    :param cleaning: bool, if True, will update the empty fields with default values +
        cleans the kwargs_fixed
    :return: updated list and kwargs
    """
    light_model_list.append("SERSIC_ELLIPSE")
    for light_name, light_param in light.parameters.items():
        if light_name == "I_eff":
            amp = getattr(light_param.point_estimate, "value")
            amp_up = getattr(light_param.definition_range, "max_value")
            amp_down = getattr(light_param.definition_range, "min_value")
            amp_fixed = amp if getattr(light_param, "fixed") else None
        elif light_name == "theta_eff":
            R = getattr(light_param.point_estimate, "value")
            R_up = getattr(light_param.definition_range, "max_value")
            R_down = getattr(light_param.definition_range, "min_value")
            R_fixed = R if getattr(light_param, "fixed") else None
        elif light_name == "n":
            n = getattr(light_param.point_estimate, "value")
            n_up = getattr(light_param.definition_range, "max_value")
            n_down = getattr(light_param.definition_range, "min_value")
            n_fixed = n if getattr(light_param, "fixed") else None
        elif light_name == "q":
            q = getattr(light_param.point_estimate, "value")
            q_up = getattr(light_param.definition_range, "max_value")
            q_down = getattr(light_param.definition_range, "min_value")
            q_fixed = q if getattr(light_param, "fixed") else None
        elif light_name == "phi":
            phi = getattr(light_param.point_estimate, "value")
            phi_up = getattr(light_param.definition_range, "max_value")
            phi_down = getattr(light_param.definition_range, "min_value")
            phi_fixed = phi if getattr(light_param, "fixed") else None
        elif light_name == "center_x":
            cx = (
                -getattr(light_param.point_estimate, "value")
                if getattr(light_param.point_estimate, "value") is not None
                else None
            )
            cx_up = getattr(light_param.definition_range, "max_value")
            cx_down = getattr(light_param.definition_range, "min_value")
            cx_fixed = cx if getattr(light_param, "fixed") else None
        elif light_name == "center_y":
            cy = getattr(light_param.point_estimate, "value")
            cy_up = getattr(light_param.definition_range, "max_value")
            cy_down = getattr(light_param.definition_range, "min_value")
            cy_fixed = cy if getattr(light_param, "fixed") else None
        else:
            print(f"Parameter {light_name} unknown in SersicEllipse profile.")

    e1, e2 = qphi_coolest_to_e1e2_lenstronomy(q, phi)
    e1_fixed, e2_fixed = qphi_coolest_to_e1e2_lenstronomy(q_fixed, phi_fixed)
    e1_down, e1_up, e2_down, e2_up = ellibounds_coolest_to_lenstronomy(
        q_down, q_up, phi_down, phi_up
    )

    kw_1 = {
        "amp": amp,
        "R_sersic": R,
        "n_sersic": n,
        "e1": e1,
        "e2": e2,
        "center_x": cx,
        "center_y": cy,
    }
    kw_up_1 = {
        "R_sersic": R_up,
        "n_sersic": n_up,
        "e1": e1_up,
        "e2": e2_up,
        "center_x": cx_up,
        "center_y": cy_up,
    }
    kw_down_1 = {
        "R_sersic": R_down,
        "n_sersic": n_down,
        "e1": e1_down,
        "e2": e2_down,
        "center_x": cx_down,
        "center_y": cy_down,
    }
    kw_fixed_1 = {
        "R_sersic": R_fixed,
        "n_sersic": n_fixed,
        "e1": e1_fixed,
        "e2": e2_fixed,
        "center_x": cx_fixed,
        "center_y": cy_fixed,
    }

    kw_ = kw_1.copy()
    kw_init = kw_1.copy()
    kw_up = kw_up_1.copy()
    kw_down = kw_down_1.copy()
    kw_fixed = kw_fixed_1.copy()

    if cleaning is True:
        kw_init_default = {
            "amp": 1.0,
            "R_sersic": 1.0,
            "n_sersic": 3.5,
            "e1": 0.0,
            "e2": -0.0,
            "center_x": 0.0,
            "center_y": 0.0,
        }
        kw_up_default = {
            "amp": 100,
            "R_sersic": 100,
            "n_sersic": 8,
            "e1": 0.5,
            "e2": 0.5,
            "center_x": 100,
            "center_y": 100,
        }
        kw_down_default = {
            "amp": 0,
            "R_sersic": 0,
            "n_sersic": 0.5,
            "e1": -0.5,
            "e2": -0.5,
            "center_x": -100,
            "center_y": -100,
        }
        for key, val in kw_1.items():
            if val is None:
                kw_init[key] = kw_init_default[key]
        for key, val in kw_up_1.items():
            if val is None:
                kw_up[key] = kw_up_default[key]
        for key, val in kw_down_1.items():
            if val is None:
                kw_down[key] = kw_down_default[key]
        for key, val in kw_fixed_1.items():
            if val is None:
                del kw_fixed[key]

    kwargs_light.append(kw_)
    kwargs_light_init.append(kw_init)
    kwargs_light_up.append(kw_up)
    kwargs_light_down.append(kw_down)
    kwargs_light_fixed.append(kw_fixed)
    kwargs_light_sigma.append(
        {
            "R_sersic": 1.5,
            "n_sersic": 1.5,
            "e1": 0.2,
            "e2": 0.2,
            "center_x": 0.5,
            "center_y": 0.5,
        }
    )
    print("\t Sersic (Ellipse) correctly added")
    return


def update_kwargs_shapelets(
    light,
    light_model_list,
    kwargs_light,
    kwargs_light_init,
    kwargs_light_up,
    kwargs_light_down,
    kwargs_light_fixed,
    kwargs_light_sigma,
    cleaning=False,
):
    """Update the source list and kwargs with SHAPELETS light model.

    :param light: coolest.template.classes.profiles.light.Shapelets object
    :param light_model_list: the usual lenstronomy lens_light_model_list or
        source_light_model_list
    :param kwargs_light: the usual lenstronomy kwargs
    :param kwargs_light_init: the usual lenstronomy kwargs
    :param kwargs_light_up: the usual lenstronomy kwargs
    :param kwargs_light_down: the usual lenstronomy kwargs
    :param kwargs_light_fixed: the usual lenstronomy kwargs
    :param kwargs_light_sigma: the usual lenstronomy kwargs
    :param cleaning: bool, if True, will update the empty fields with default values +
        cleans the kwargs_fixed
    :return: updated list and kwargs
    """
    light_model_list.append("SHAPELETS")
    for light_name, light_param in light.parameters.items():
        if light_name == "beta":
            b = getattr(light_param.point_estimate, "value")
            b_up = getattr(light_param.definition_range, "max_value")
            b_down = getattr(light_param.definition_range, "min_value")
            b_fixed = b if getattr(light_param, "fixed") else None
        elif light_name == "n_max":
            nmax = getattr(light_param.point_estimate, "value")
            nmax_fixed = nmax if getattr(light_param, "fixed") else None
        elif light_name == "center_x":
            cx = (
                -getattr(light_param.point_estimate, "value")
                if getattr(light_param.point_estimate, "value") is not None
                else None
            )
            cx_up = getattr(light_param.definition_range, "max_value")
            cx_down = getattr(light_param.definition_range, "min_value")
            cx_fixed = cx if getattr(light_param, "fixed") else None
        elif light_name == "center_y":
            cy = getattr(light_param.point_estimate, "value")
            cy_up = getattr(light_param.definition_range, "max_value")
            cy_down = getattr(light_param.definition_range, "min_value")
            cy_fixed = cy if getattr(light_param, "fixed") else None
        elif light_name == "amps":
            amp = shapelet_amp_coolest_to_lenstronomy(
                getattr(light_param.point_estimate, "value")
            )
            amp_up = shapelet_amp_coolest_to_lenstronomy(
                getattr(light_param.definition_range, "max_value")
            )
            amp_down = shapelet_amp_coolest_to_lenstronomy(
                getattr(light_param.definition_range, "min_value")
            )
            amp_fixed = amp if getattr(light_param, "fixed") else None
        else:
            print(f"Parameter {light_name} unknown in Shapelets profile.")

    kw_1 = {"amp": amp, "beta": b, "center_x": cx, "center_y": cy, "n_max": nmax}
    kw_up_1 = {"beta": b_up, "center_x": cx_up, "center_y": cy_up}
    kw_down_1 = {"beta": b_down, "center_x": cx_down, "center_y": cy_down}
    kw_fixed_1 = {
        "beta": b_fixed,
        "center_x": cx_fixed,
        "center_y": cy_fixed,
        "n_max": nmax,
    }

    kw_ = kw_1.copy()
    kw_init = kw_1.copy()
    kw_up = kw_up_1.copy()
    kw_down = kw_down_1.copy()
    kw_fixed = kw_fixed_1.copy()

    if cleaning is True:
        kw_init_default = {
            "amp": 1.0,
            "beta": 0.1,
            "center_x": 0.0,
            "center_y": 0.0,
            "n_max": -1,
        }
        kw_up_default = {
            "amp": 100,
            "beta": 100,
            "center_x": 100,
            "center_y": 100,
            "n_max": 1000,
        }
        kw_down_default = {
            "amp": 0,
            "beta": 0,
            "center_x": -100,
            "center_y": -100,
            "n_max": -1,
        }
        for key, val in kw_1.items():
            if val is None:
                kw_init[key] = kw_init_default[key]
        for key, val in kw_up_1.items():
            if val is None:
                kw_up[key] = kw_up_default[key]
        for key, val in kw_down_1.items():
            if val is None:
                kw_down[key] = kw_down_default[key]
        for key, val in kw_fixed_1.items():
            if val is None:
                del kw_fixed[key]

    kwargs_light.append(kw_)
    kwargs_light_init.append(kw_init)
    kwargs_light_up.append(kw_up)
    kwargs_light_down.append(kw_down)
    kwargs_light_fixed.append(kw_fixed)
    kwargs_light_sigma.append({"beta": 0.5, "center_x": 0.5, "center_y": 0.5})
    print("\t Shapelets correctly added")
    return


def update_kwargs_lensed_ps(
    light,
    ps_model_list,
    kwargs_ps,
    kwargs_ps_init,
    kwargs_ps_up,
    kwargs_ps_down,
    kwargs_ps_fixed,
    kwargs_ps_sigma,
    cleaning=False,
):
    """Update the source list and kwargs with lensed point source "LENSED_POSITION"
    light model.

    :param light: coolest.template.classes.profiles.lightLensedPS object
    :param ps_model_list: the usual lenstronomy point_source_model_list
    :param kwargs_ps: the usual lenstronomy kwargs
    :param kwargs_ps_init: the usual lenstronomy kwargs
    :param kwargs_ps_up: the usual lenstronomy kwargs
    :param kwargs_ps_down: the usual lenstronomy kwargs
    :param kwargs_ps_fixed: the usual lenstronomy kwargs
    :param kwargs_ps_sigma: the usual lenstronomy kwargs
    :param cleaning: bool, if True, will update the empty fields with default values +
        cleans the kwargs_fixed
    :return: updated list and kwargs
    """
    ps_model_list.append("LENSED_POSITION")

    try:
        num_ps = len(getattr(light.parameters["ra_list"].point_estimate, "value"))
    except:
        num_ps = 4

    for light_name, light_param in light.parameters.items():
        if light_name == "ra_list":
            ra = (
                -np.array(getattr(light_param.point_estimate, "value"))
                if getattr(light_param.point_estimate, "value") is not None
                else None
            )
            ra_up = getattr(light_param.definition_range, "max_value")
            ra_down = getattr(light_param.definition_range, "min_value")
            ra_fixed = ra if getattr(light_param, "fixed") else None
        elif light_name == "dec_list":
            dec = getattr(light_param.point_estimate, "value")
            dec_up = getattr(light_param.definition_range, "max_value")
            dec_down = getattr(light_param.definition_range, "min_value")
            dec_fixed = dec if getattr(light_param, "fixed") else None
        elif light_name == "amps":
            amp = getattr(light_param.point_estimate, "value")
            amp_up = getattr(light_param.definition_range, "max_value")
            amp_down = getattr(light_param.definition_range, "min_value")
            amp_fixed = amp if getattr(light_param, "fixed") else None
        else:
            print(f"Parameter {light_name} unknown in LensedPS profile.")

    kw_1 = {
        "point_amp": np.array(amp) if amp is not None else None,
        "ra_image": np.array(ra) if ra is not None else None,
        "dec_image": np.array(dec) if dec is not None else None,
    }
    kw_up_1 = {
        "ra_image": np.array(ra_up) if ra_up is not None else None,
        "dec_image": np.array(dec_up) if dec_up is not None else None,
    }
    kw_down_1 = {
        "ra_image": np.array(ra_down) if ra_down is not None else None,
        "dec_image": np.array(dec_down) if dec_down is not None else None,
    }
    kw_fixed_1 = {"ra_image": ra_fixed, "dec_image": dec_fixed}

    kw_ = kw_1.copy()
    kw_init = kw_1.copy()
    kw_up = kw_up_1.copy()
    kw_down = kw_down_1.copy()
    kw_fixed = kw_fixed_1.copy()

    if cleaning is True:
        kw_init_default = {
            "point_amp": np.ones(num_ps),
            "ra_image": np.ones(num_ps),
            "dec_image": np.ones(num_ps),
        }
        kw_up_default = {
            "ra_image": np.ones(num_ps) * 10,
            "dec_image": np.ones(num_ps) * 10,
        }
        kw_down_default = {
            "ra_image": np.ones(num_ps) * (-10),
            "dec_image": np.ones(num_ps) * (-10),
        }
        for key, val in kw_1.items():
            if val is None:
                kw_init[key] = kw_init_default[key]
        for key, val in kw_up_1.items():
            if val is None:
                kw_up[key] = kw_up_default[key]
        for key, val in kw_down_1.items():
            if val is None:
                kw_down[key] = kw_down_default[key]
        for key, val in kw_fixed_1.items():
            if val is None:
                del kw_fixed[key]

    kwargs_ps.append(kw_)
    kwargs_ps_init.append(kw_init)
    kwargs_ps_up.append(kw_up)
    kwargs_ps_down.append(kw_down)
    kwargs_ps_fixed.append(kw_fixed)
    kwargs_ps_sigma.append({"ra_image": np.ones(num_ps), "dec_image": np.ones(num_ps)})
    print("\t Lensed point sources correctly added")
    return



import astropy
from scipy.integrate import quad

if float(astropy.__version__[0]) < 5.0:
    Warning(
        "This routines are only supported for astropy version >=5. Current version is %s."
        % astropy.__version__
    )
else:
    from astropy.cosmology.utils import vectorize_redshift_method


class CosmoInterp(object):
    """Class which interpolates the comoving transfer distance and then computes angular
    diameter distances from it This class is modifying the astropy.cosmology
    routines."""

    def __init__(self, cosmo):
        """

        :param cosmo: astropy.cosmology instance (version 4.0 as private functions need to be supported)
        """
        self._cosmo = cosmo

    def _integral_comoving_distance_z1z2(self, z1, z2):
        """Comoving line-of-sight distance in Mpc between objects at redshifts ``z1``
        and ``z2``. The comoving distance along the line-of-sight between two objects
        remains constant with time for objects in the Hubble flow.

        Parameters
        ----------
        z1, z2 : Quantity-like ['redshift'] or array-like
            Input redshifts.

        Returns
        -------
        d : `~astropy.units.Quantity` ['length']
            Comoving distance in Mpc between each input redshift.
        """

        return (
            self._cosmo._hubble_distance
            * self._integral_comoving_distance_z1z2_scalar(z1, z2)
        )

    @vectorize_redshift_method(nin=2)
    def _integral_comoving_distance_z1z2_scalar(self, z1, z2):
        """Comoving line-of-sight distance between objects at redshifts ``z1`` and
        ``z2``. Value in Mpc.

        The comoving distance along the line-of-sight between two objects
        remains constant with time for objects in the Hubble flow.

        Parameters
        ----------
        z1, z2 : Quantity-like ['redshift'], array-like, or `~numbers.Number`
            Input redshifts.

        Returns
        -------
        d : float or ndarray
            Comoving distance in Mpc between each input redshift.
            Returns `float` if input scalar, `~numpy.ndarray` otherwise.
        """
        return quad(
            self._cosmo._inv_efunc_scalar,
            z1,
            z2,
            args=self._cosmo._inv_efunc_scalar_args,
        )[0]



__author__ = "sibirrer"

import numpy.testing as npt
import numpy as np
import pytest
from lenstronomy.LensModel.Profiles.curved_arc_tan_diff import CurvedArcTanDiff
from lenstronomy.LensModel.Profiles.sie import SIE
from lenstronomy.LensModel.Profiles.convergence import Convergence
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.lens_model_extensions import LensModelExtensions


class TestCurvedArcTanDiff(object):
    """Tests the source model routines."""

    def setup_method(self):
        self.model = CurvedArcTanDiff()
        self.sie = SIE()
        self.mst = Convergence()

    def test_curved_arc_round_recovery(self):
        """Test whether the curved arc parameters are satisfied in differential form."""

        center_x, center_y = (
            2,
            0.0,
        )  # test works except at (0,0) where the direction angle is not well defined
        tangential_stretch = 4.0
        radial_stretch = 1.0
        curvature, direction = 0.5, 0.5
        dtan_dtan = 0
        kwargs_lens = {
            "tangential_stretch": tangential_stretch,
            "radial_stretch": radial_stretch,
            "curvature": curvature,
            "direction": direction,
            "dtan_dtan": dtan_dtan,
            "center_x": center_x,
            "center_y": center_y,
        }
        self._test_curved_arc_recovery(kwargs_lens)

        kwargs_lens = {
            "tangential_stretch": tangential_stretch,
            "radial_stretch": 1.1,
            "curvature": curvature,
            "direction": 0.01,
            "dtan_dtan": dtan_dtan,
            "center_x": center_x,
            "center_y": center_y,
        }
        self._test_curved_arc_recovery(kwargs_lens)

        kwargs_lens = {
            "tangential_stretch": 10,
            "radial_stretch": 1.0,
            "curvature": 0.2,
            "direction": 0.01,
            "dtan_dtan": 0.0,
            "center_x": center_x,
            "center_y": center_y,
        }
        self._test_curved_arc_recovery(kwargs_lens)

    def test_curved_arc_recovery(self):
        """Test whether the curved arc parameters are satisfied in differential form."""

        center_x, center_y = (
            3,
            0,
        )  # test works except at (0,0) where the direction angle is not well defined

        kwargs_lens = {
            "tangential_stretch": 2.0,
            "radial_stretch": 1.0,
            "curvature": 0.3,
            "direction": 0.001,
            "dtan_dtan": 0.1,
            "center_x": center_x,
            "center_y": center_y,
        }
        self._test_curved_arc_recovery(kwargs_lens)

        # and here we change directions
        kwargs_lens = {
            "tangential_stretch": 2.0,
            "radial_stretch": 1.0,
            "curvature": 0.3,
            "direction": 0.5,
            "dtan_dtan": 0.1,
            "center_x": center_x,
            "center_y": center_y,
        }
        self._test_curved_arc_recovery(kwargs_lens)

        # and here we have the radial stretch != 1, thus applying an MST
        kwargs_lens = {
            "tangential_stretch": 2.0,
            "radial_stretch": 1.1,
            "curvature": 0.3,
            "direction": 0.5,
            "dtan_dtan": 0.1,
            "center_x": center_x,
            "center_y": center_y,
        }
        self._test_curved_arc_recovery(kwargs_lens)

        kwargs_lens = {
            "tangential_stretch": 2.0,
            "radial_stretch": 1.1,
            "curvature": 0.3,
            "direction": 0.5,
            "dtan_dtan": -0.1,
            "center_x": center_x,
            "center_y": center_y,
        }
        self._test_curved_arc_recovery(kwargs_lens)

    def _test_in_out_scaling(self):
        # some scaling tests with plots that are going to be ignored
        ext = LensModelExtensions(LensModel(lens_model_list=["CURVED_ARC_TAN_DIFF"]))

        # change in dtan_dtan for fixed other components input vs output comparison
        dtan_dtan_list = np.linspace(start=-0.5, stop=0.5, num=21)
        dtan_dtan_out_list = []
        kwargs_lens = {
            "tangential_stretch": 2.0,
            "radial_stretch": 1.0,
            "curvature": 0.2,
            "direction": 0.001,
            "dtan_dtan": 0.1,
            "center_x": 1,
            "center_y": 0,
        }
        for dtan_dtan in dtan_dtan_list:
            kwargs_lens["dtan_dtan"] = dtan_dtan
            (
                lambda_rad,
                lambda_tan,
                orientation_angle,
                dlambda_tan_dtan,
                dlambda_tan_drad,
                dlambda_rad_drad,
                dlambda_rad_dtan,
                dphi_tan_dtan,
                dphi_tan_drad,
                dphi_rad_drad,
                dphi_rad_dtan,
            ) = ext.radial_tangential_differentials(1, 0, [kwargs_lens])
            dtan_dtan_out_list.append(dlambda_tan_dtan)
        dtan_dtan_out_list = np.array(dtan_dtan_out_list)
        import matplotlib.pyplot as plt

        plt.plot(dtan_dtan_list, dtan_dtan_out_list)
        plt.xlabel("dtan_in fixed lens")
        plt.ylabel("dtan_out")
        plt.show()
        # npt.assert_almost_equal(dtan_dtan_out_list, dtan_dtan_list)

        # change in tangential stretch
        lambda_tan_list = np.linspace(start=2, stop=20, num=21)
        dtan_dtan_out_list = []
        kwargs_lens = {
            "tangential_stretch": 2.0,
            "radial_stretch": 1.0,
            "curvature": 0.2,
            "direction": 0.001,
            "dtan_dtan": 0.1,
            "center_x": 1,
            "center_y": 0,
        }
        for lambda_tan in lambda_tan_list:
            kwargs_lens["tangential_stretch"] = lambda_tan
            (
                lambda_rad,
                lambda_tan,
                orientation_angle,
                dlambda_tan_dtan,
                dlambda_tan_drad,
                dlambda_rad_drad,
                dlambda_rad_dtan,
                dphi_tan_dtan,
                dphi_tan_drad,
                dphi_rad_drad,
                dphi_rad_dtan,
            ) = ext.radial_tangential_differentials(1, 0, [kwargs_lens])
            dtan_dtan_out_list.append(dlambda_tan_dtan)
        dtan_dtan_out_list = np.array(dtan_dtan_out_list)
        import matplotlib.pyplot as plt

        plt.plot(lambda_tan_list, dtan_dtan_out_list)
        plt.xlabel("lambda_tan")
        plt.ylabel("dtan_out")
        plt.show()
        # npt.assert_almost_equal(dtan_dtan_out_list, dtan_dtan_list)

        # change in curvature radius
        curvature_list = np.linspace(start=0.1, stop=1, num=21)
        dtan_dtan_out_list = []
        kwargs_lens = {
            "tangential_stretch": 5.0,
            "radial_stretch": 1.0,
            "curvature": 0.2,
            "direction": 0.001,
            "dtan_dtan": 0.1,
            "center_x": 1,
            "center_y": 0,
        }
        for curvatrue in curvature_list:
            kwargs_lens["curvature"] = curvatrue
            (
                lambda_rad,
                lambda_tan,
                orientation_angle,
                dlambda_tan_dtan,
                dlambda_tan_drad,
                dlambda_rad_drad,
                dlambda_rad_dtan,
                dphi_tan_dtan,
                dphi_tan_drad,
                dphi_rad_drad,
                dphi_rad_dtan,
            ) = ext.radial_tangential_differentials(1, 0, [kwargs_lens])
            dtan_dtan_out_list.append(dlambda_tan_dtan)
        dtan_dtan_out_list = np.array(dtan_dtan_out_list)
        import matplotlib.pyplot as plt

        plt.plot(lambda_tan_list, dtan_dtan_out_list)
        plt.xlabel("curvature")
        plt.ylabel("dtan_out")
        plt.show()
        # npt.assert_almost_equal(dtan_dtan_out_list, dtan_dtan_list)

    def _test_curved_arc_recovery(self, kwargs_arc_init):
        ext = LensModelExtensions(LensModel(lens_model_list=["CURVED_ARC_TAN_DIFF"]))
        center_x, center_y = kwargs_arc_init["center_x"], kwargs_arc_init["center_y"]
        kwargs_arc = ext.curved_arc_estimate(center_x, center_y, [kwargs_arc_init])
        (
            lambda_rad,
            lambda_tan,
            orientation_angle,
            dlambda_tan_dtan,
            dlambda_tan_drad,
            dlambda_rad_drad,
            dlambda_rad_dtan,
            dphi_tan_dtan,
            dphi_tan_drad,
            dphi_rad_drad,
            dphi_rad_dtan,
        ) = ext.radial_tangential_differentials(center_x, center_y, [kwargs_arc_init])
        print(lambda_tan, dlambda_tan_dtan, kwargs_arc_init["dtan_dtan"])
        npt.assert_almost_equal(
            kwargs_arc["tangential_stretch"] / kwargs_arc_init["tangential_stretch"],
            1,
            decimal=3,
        )
        npt.assert_almost_equal(
            kwargs_arc["radial_stretch"], kwargs_arc_init["radial_stretch"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_arc["curvature"], kwargs_arc_init["curvature"], decimal=2
        )
        npt.assert_almost_equal(dphi_tan_dtan, kwargs_arc_init["curvature"], decimal=2)
        npt.assert_almost_equal(
            kwargs_arc["direction"], kwargs_arc_init["direction"], decimal=3
        )
        npt.assert_almost_equal(
            dlambda_tan_dtan / lambda_tan, kwargs_arc_init["dtan_dtan"], decimal=2
        )


if __name__ == "__main__":
    pytest.main("-k TestLensModel")



__author__ = "sibirrer"

from lenstronomy.Sampling.Likelihoods.time_delay_likelihood import TimeDelayLikelihood
from lenstronomy.Sampling.Likelihoods.image_likelihood import ImageLikelihood
from lenstronomy.Sampling.Likelihoods.position_likelihood import PositionLikelihood
from lenstronomy.Sampling.Likelihoods.flux_ratio_likelihood import FluxRatioLikelihood
from lenstronomy.Sampling.Likelihoods.prior_likelihood import PriorLikelihood
from lenstronomy.Sampling.Likelihoods.kinematic_2D_likelihood import KinLikelihood
import lenstronomy.Util.class_creator as class_creator
import numpy as np

__all__ = ["LikelihoodModule"]


class LikelihoodModule(object):
    """This class contains the routines to run a MCMC process.

    the key components are:
    - imSim_class: an instance of a class that simulates one (or more) images and returns the likelihood, such as
    ImageModel(), Multiband(), MultiExposure()
    - param_class: instance of a Param() class that can cast the sorted list of parameters that are sampled into the
    conventions of the imSim_class

    Additional arguments are supported for adding a time-delay likelihood etc (see __init__ definition)
    """

    def __init__(
        self,
        kwargs_data_joint,
        kwargs_model,
        param_class,
        image_likelihood=True,
        check_bounds=True,
        check_matched_source_position=False,
        astrometric_likelihood=False,
        image_position_likelihood=False,
        source_position_likelihood=False,
        image_position_uncertainty=0.004,
        check_positive_flux=False,
        source_position_tolerance=0.001,
        source_position_sigma=0.001,
        force_no_add_image=False,
        source_marg=False,
        linear_prior=None,
        restrict_image_number=False,
        max_num_images=None,
        bands_compute=None,
        time_delay_likelihood=False,
        image_likelihood_mask_list=None,
        flux_ratio_likelihood=False,
        kwargs_flux_compute=None,
        prior_lens=None,
        prior_source=None,
        prior_extinction=None,
        prior_lens_light=None,
        prior_ps=None,
        prior_special=None,
        prior_lens_kde=None,
        prior_source_kde=None,
        prior_lens_light_kde=None,
        prior_ps_kde=None,
        prior_special_kde=None,
        prior_extinction_kde=None,
        prior_lens_lognormal=None,
        prior_source_lognormal=None,
        prior_extinction_lognormal=None,
        prior_lens_light_lognormal=None,
        prior_ps_lognormal=None,
        prior_special_lognormal=None,
        custom_logL_addition=None,
        kwargs_pixelbased=None,
        kinematic_2d_likelihood=False,
        kin_lens_idx=0,
        kin_lens_light_idx=0,
    ):
        """Initializing class.

        :param param_class: instance of a Param() class that can cast the sorted list of
            parameters that are sampled into the conventions of the imSim_class
        :param image_likelihood: bool, option to compute the imaging likelihood
        :param source_position_likelihood: bool, if True, ray-traces image positions
            back to source plane and evaluates relative errors in respect ot the
            position_uncertainties in the image plane
        :param check_bounds: bool, option to punish the hard bounds in parameter space
        :param check_matched_source_position: bool, option to check whether point source
            position of solver finds a solution to match all the image positions in the
            same source plane coordinate
        :param astrometric_likelihood: bool, additional likelihood term of the predicted
            vs modelled point source position
        :param image_position_uncertainty: float, 1-sigma Gaussian uncertainty on the
            point source position (only used if point_source_likelihood=True)
        :param check_positive_flux: bool, option to punish models that do not have all
            positive linear amplitude parameters
        :param source_position_tolerance: float, punishment of check_solver occurs when
            image positions are predicted further away than this number
        :param image_likelihood_mask_list: list of boolean 2d arrays of size of images
            marking the pixels to be evaluated in the likelihood
        :param force_no_add_image: bool, if True: computes ALL image positions of the
            point source. If there are more images predicted than modelled, a punishment
            occurs
        :param source_marg: marginalization addition on the imaging likelihood based on
            the covariance of the inferred linear coefficients
        :param linear_prior: float or list of floats (when multi-linear setting is
            chosen) indicating the range of linear amplitude priors when computing the
            marginalization term.
        :param restrict_image_number: bool, if True: computes ALL image positions of the
            point source. If there are more images predicted than indicated in
            max_num_images, a punishment occurs
        :param max_num_images: int, see restrict_image_number
        :param bands_compute: list of bools with same length as data objects, indicates
            which "band" to include in the fitting
        :param time_delay_likelihood: bool, if True computes the time-delay likelihood
            of the FIRST point source
        :param kwargs_flux_compute: keyword arguments of how to compute the image
            position fluxes (see FluxRatioLikeliood)
        :param custom_logL_addition: a definition taking as arguments (kwargs_lens,
            kwargs_source, kwargs_lens_light, kwargs_ps, kwargs_special,
            kwargs_extinction) and returns a logL (punishing) value.
        :param kwargs_pixelbased: keyword arguments with various settings related to the
            pixel-based solver (see SLITronomy documentation)
        :param kinematic_2d_likelihood: bool, option to compute the kinematic likelihood
        """
        (
            multi_band_list,
            multi_band_type,
            time_delays_measured,
            time_delays_uncertainties,
            flux_ratios,
            flux_ratio_errors,
            ra_image_list,
            dec_image_list,
            kinematic_data,
        ) = self._unpack_data(**kwargs_data_joint)
        if len(multi_band_list) == 0:
            image_likelihood = False
        self.kinematic_data = kinematic_data
        self.param = param_class
        self._lower_limit, self._upper_limit = self.param.param_limits()
        self._prior_likelihood = PriorLikelihood(
            prior_lens,
            prior_source,
            prior_lens_light,
            prior_ps,
            prior_special,
            prior_extinction,
            prior_lens_kde,
            prior_source_kde,
            prior_lens_light_kde,
            prior_ps_kde,
            prior_special_kde,
            prior_extinction_kde,
            prior_lens_lognormal,
            prior_source_lognormal,
            prior_lens_light_lognormal,
            prior_ps_lognormal,
            prior_special_lognormal,
            prior_extinction_lognormal,
        )
        self._time_delay_likelihood = time_delay_likelihood
        self._image_likelihood = image_likelihood
        self._flux_ratio_likelihood = flux_ratio_likelihood
        self._kinematic_2D_likelihood = kinematic_2d_likelihood
        if kwargs_flux_compute is None:
            kwargs_flux_compute = {}
        linear_solver = self.param.linear_solver
        self._kwargs_flux_compute = kwargs_flux_compute
        self._check_bounds = check_bounds
        self._custom_logL_addition = custom_logL_addition
        self._kwargs_time_delay = {
            "time_delays_measured": time_delays_measured,
            "time_delays_uncertainties": time_delays_uncertainties,
        }
        self.kwargs_imaging = {
            "multi_band_list": multi_band_list,
            "multi_band_type": multi_band_type,
            "bands_compute": bands_compute,
            "image_likelihood_mask_list": image_likelihood_mask_list,
            "source_marg": source_marg,
            "linear_prior": linear_prior,
            "check_positive_flux": check_positive_flux,
            "kwargs_pixelbased": kwargs_pixelbased,
            "linear_solver": linear_solver,
        }
        self._kwargs_position = {
            "astrometric_likelihood": astrometric_likelihood,
            "image_position_likelihood": image_position_likelihood,
            "source_position_likelihood": source_position_likelihood,
            "ra_image_list": ra_image_list,
            "dec_image_list": dec_image_list,
            "image_position_uncertainty": image_position_uncertainty,
            "check_matched_source_position": check_matched_source_position,
            "source_position_tolerance": source_position_tolerance,
            "source_position_sigma": source_position_sigma,
            "force_no_add_image": force_no_add_image,
            "restrict_image_number": restrict_image_number,
            "max_num_images": max_num_images,
        }
        self._kwargs_flux = {
            "flux_ratios": flux_ratios,
            "flux_ratio_errors": flux_ratio_errors,
        }
        self._kwargs_flux.update(self._kwargs_flux_compute)

        if self._kinematic_2D_likelihood is True:
            print(
                "Note that the 2D kinematic likelihood assumes that the lens and lens light have the same center and orientation"
            )
            if len(multi_band_list) > 1:
                print(
                    "Kinematic Likelihood not meant for multiband, using first band by default"
                )
            if kwargs_model["lens_model_list"][kin_lens_idx] not in ["EPL_Q_PHI"]:
                print(
                    "Lens for kinematic is not EPL_Q_PHI, the 2D kinematic likelihood will break."
                )
            if kwargs_model["lens_light_model_list"][kin_lens_light_idx] not in [
                "SERSIC_ELLIPSE_Q_PHI"
            ]:
                print(
                    "Lens light for kinematic is not SERSIC_ELLIPSE_Q_PHI , the 2D kinematic likelihood will break."
                )
            self._kin_lens_idx = kin_lens_idx
            self._kin_lens_light_idx = kin_lens_light_idx

        self._class_instances(
            kwargs_model=kwargs_model,
            kwargs_imaging=self.kwargs_imaging,
            kwargs_position=self._kwargs_position,
            kwargs_flux=self._kwargs_flux,
            kwargs_time_delay=self._kwargs_time_delay,
            kinematic_data=self.kinematic_data,
        )

    def _class_instances(
        self,
        kwargs_model,
        kwargs_imaging,
        kwargs_position,
        kwargs_flux,
        kwargs_time_delay,
        kinematic_data,
    ):
        """

        :param kwargs_model: lenstronomy model keyword arguments
        :param kwargs_imaging: keyword arguments for imaging likelihood
        :param kwargs_position: keyword arguments for positional likelihood
        :param kwargs_flux: keyword arguments for flux ratio likelihood
        :param kwargs_time_delay: keyword arguments for time delay likelihood
        :param kinematic_data: kinematic class for kinematic likelihood
        :return: updated model instances of this class
        """

        # TODO: in case lens model or point source models are only applied on partial images, then this current class
        # has ambiguities when it comes to position likelihood, time-delay likelihood and flux ratio likelihood
        (
            lens_model_class,
            _,
            lens_light_model_class,
            point_source_class,
            _,
        ) = class_creator.create_class_instances(all_models=True, **kwargs_model)
        self.PointSource = point_source_class

        if self._time_delay_likelihood is True:
            self.time_delay_likelihood = TimeDelayLikelihood(
                lens_model_class=lens_model_class,
                point_source_class=point_source_class,
                **kwargs_time_delay
            )

        if self._image_likelihood is True:
            self.image_likelihood = ImageLikelihood(
                kwargs_model=kwargs_model, **kwargs_imaging
            )
        self._position_likelihood = PositionLikelihood(
            point_source_class, **kwargs_position
        )
        if self._flux_ratio_likelihood is True:
            self.flux_ratio_likelihood = FluxRatioLikelihood(
                lens_model_class, **kwargs_flux
            )
        if self._kinematic_2D_likelihood is True:
            self.kinematic_2D_likelihood = KinLikelihood(
                kinematic_data,
                lens_model_class,
                lens_light_model_class,
                kwargs_imaging["multi_band_list"][0][0],
                self._kin_lens_idx,
                self._kin_lens_light_idx,
            )

    def __call__(self, a):
        return self.logL(a)

    def logL(self, args, verbose=False):
        """Routine to compute X2 given variable parameters for a MCMC/PSO chain.

        :param args: ordered parameter values that are being sampled
        :type args: tuple or list of floats
        :param verbose: if True, makes print statements about individual likelihood
            components
        :type verbose: boolean
        :returns: log likelihood of the data given the model (natural logarithm)
        """
        # extract parameters
        kwargs_return = self.param.args2kwargs(args)
        if self._check_bounds is True:
            penalty, bound_hit = self.check_bounds(
                args, self._lower_limit, self._upper_limit, verbose=verbose
            )
            if bound_hit is True:
                return -(10**15)
        return self.log_likelihood(kwargs_return, verbose=verbose)

    def log_likelihood(self, kwargs_return, verbose=False):
        """


        :param kwargs_return: need to contain 'kwargs_lens', 'kwargs_source', 'kwargs_lens_light', 'kwargs_ps',
         'kwargs_special'. These entries themselves are lists of keyword argument of the parameters entering the model
         to be evaluated
        :type kwargs_return: keyword arguments
        :param verbose: if True, makes print statements about individual likelihood components
        :type verbose: boolean

        :returns:
         - logL (float) log likelihood of the data given the model (natural logarithm)
        """
        kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps, kwargs_special = (
            kwargs_return["kwargs_lens"],
            kwargs_return["kwargs_source"],
            kwargs_return["kwargs_lens_light"],
            kwargs_return["kwargs_ps"],
            kwargs_return["kwargs_special"],
        )
        # update model instance in case of changes affecting it (i.e. redshift sampling in multi-plane)
        self._update_model(kwargs_special)
        # generate image and computes likelihood
        self._reset_point_source_cache(bool_input=True)
        logL = 0

        if self._image_likelihood is True:
            logL_image = self.image_likelihood.logL(**kwargs_return)
            logL += logL_image
            if verbose is True:
                print("image logL = %s" % logL_image)
        if self._time_delay_likelihood is True:
            logL_time_delay = self.time_delay_likelihood.logL(
                kwargs_lens, kwargs_ps, kwargs_special
            )
            logL += logL_time_delay
            if verbose is True:
                print("time-delay logL = %s" % logL_time_delay)
        if self._flux_ratio_likelihood is True:
            ra_image_list, dec_image_list = self.PointSource.image_position(
                kwargs_ps=kwargs_ps, kwargs_lens=kwargs_lens
            )
            x_pos, y_pos = ra_image_list[0], dec_image_list[0]
            logL_flux_ratios = self.flux_ratio_likelihood.logL(
                x_pos, y_pos, kwargs_lens, kwargs_special
            )
            logL += logL_flux_ratios
            if verbose is True:
                print("flux ratio logL = %s" % logL_flux_ratios)
        if self._kinematic_2D_likelihood is True:
            logL_kinematic_2d = self.kinematic_2D_likelihood.logL(
                kwargs_lens, kwargs_lens_light, kwargs_special
            )
            logL += logL_kinematic_2d
            if verbose is True:
                print("kinematic logL = %s" % logL_kinematic_2d)
        logL += self._position_likelihood.logL(
            kwargs_lens, kwargs_ps, kwargs_special, verbose=verbose
        )
        logL_prior = self._prior_likelihood.logL(**kwargs_return)
        logL += logL_prior
        if verbose is True:
            print("Prior likelihood = %s" % logL_prior)
        if self._custom_logL_addition is not None:
            logL_cond = self._custom_logL_addition(**kwargs_return)
            logL += logL_cond
            if verbose is True:
                print("custom added logL = %s" % logL_cond)
        self._reset_point_source_cache(bool_input=False)
        return logL  # , None

    @staticmethod
    def check_bounds(args, lowerLimit, upperLimit, verbose=False):
        """Checks whether the parameter vector has left its bound, if so, adds a big
        number."""
        penalty = 0.0
        bound_hit = False
        args = np.atleast_1d(args)
        for i in range(0, len(args)):
            if args[i] < lowerLimit[i] or args[i] > upperLimit[i]:
                penalty = 10.0**5
                bound_hit = True
                if verbose is True:
                    print(
                        "parameter %s with value %s hit the bounds [%s, %s] "
                        % (i, args[i], lowerLimit[i], upperLimit[i])
                    )
                return penalty, bound_hit
        return penalty, bound_hit

    @property
    def num_data(self):
        """

        :return: number of independent data points in the combined fitting
        """
        num_data = 0
        if self._image_likelihood is True:
            num_data += self.image_likelihood.num_data
        if self._time_delay_likelihood is True:
            num_data += self.time_delay_likelihood.num_data
        if self._flux_ratio_likelihood is True:
            num_data += self.flux_ratio_likelihood.num_data
        num_data += self._position_likelihood.num_data
        return num_data

    @property
    def param_limits(self):
        return self._lower_limit, self._upper_limit

    def effective_num_data_points(self, **kwargs):
        """Returns the effective number of data points considered in the X2 estimation
        to compute the reduced X2 value."""
        num_linear = 0
        if self._image_likelihood is True:
            num_linear = self.image_likelihood.num_param_linear(**kwargs)
        num_param, param_names = self.param.num_param()
        return self.num_data - num_param - num_linear

    def likelihood(self, a):
        return self.logL(a)

    def negativelogL(self, a):
        """For minimizer function, the negative value of the logl value is requested.

        :param a: array of parameters
        :return: -logL
        """
        return -self.logL(a)

    @staticmethod
    def _unpack_data(
        multi_band_list=None,
        multi_band_type="multi-linear",
        time_delays_measured=None,
        time_delays_uncertainties=None,
        flux_ratios=None,
        flux_ratio_errors=None,
        ra_image_list=None,
        dec_image_list=None,
        kinematic_data=None,
    ):
        """

        :param multi_band_list: list of [[kwargs_data, kwargs_psf, kwargs_numerics], [], ...]
        :param multi_band_type: string, type of multi-plane settings (multi-linear or joint-linear)
        :param time_delays_measured: measured time delays (units of days)
        :param time_delays_uncertainties: uncertainties in time-delay measurement
        :param flux_ratios: flux ratios of point sources
        :param flux_ratio_errors: error in flux ratio measurement
        :return:
        """
        if multi_band_list is None:
            multi_band_list = []
        if ra_image_list is None:
            ra_image_list = []
        if dec_image_list is None:
            dec_image_list = []
        return (
            multi_band_list,
            multi_band_type,
            time_delays_measured,
            time_delays_uncertainties,
            flux_ratios,
            flux_ratio_errors,
            ra_image_list,
            dec_image_list,
            kinematic_data,
        )

    def _reset_point_source_cache(self, bool_input=True):
        self.PointSource.delete_lens_model_cache()
        self.PointSource.set_save_cache(bool_input)
        if self._image_likelihood is True:
            self.image_likelihood.reset_point_source_cache(bool_input)

    def _update_model(self, kwargs_special):
        """Updates lens model instance of this class (and all class instances related to
        it) when an update to the modeled redshifts of the deflector and/or source
        planes are made.

        :param kwargs_special: keyword arguments from SpecialParam() class return of
            sampling arguments
        :return: None, all class instances updated to recent model
        """
        kwargs_model, update_bool = self.param.update_kwargs_model(kwargs_special)
        if update_bool is True:
            self._class_instances(
                kwargs_model=kwargs_model,
                kwargs_imaging=self.kwargs_imaging,
                kwargs_position=self._kwargs_position,
                kwargs_flux=self._kwargs_flux,
                kwargs_time_delay=self._kwargs_time_delay,
                kinematic_data=self.kinematic_data,
            )
        # TODO remove redundancies with Param() calls updates



import pytest
import numpy as np
import numpy.testing as npt
import copy
import unittest

from lenstronomy.Data.imaging_data import ImageData
import lenstronomy.Util.util as util


class TestData(object):
    def setup_method(self):
        self.numPix = 10
        kwargs_data = {
            "image_data": np.zeros((self.numPix, self.numPix)),
            "noise_map": np.ones((self.numPix, self.numPix)),
        }
        self.Data = ImageData(**kwargs_data)

    def test_numData(self):
        assert self.Data.num_pixel == self.numPix**2

    def test_shift_coords(self):
        numPix = 10
        deltaPix = 0.05
        (
            x_grid,
            y_grid,
            ra_at_xy_0,
            dec_at_xy_0,
            x_at_radec_0,
            y_at_radec_0,
            Mpix2coord,
            Mcoord2pix,
        ) = util.make_grid_with_coordtransform(
            numPix=numPix, deltapix=deltaPix, subgrid_res=1, inverse=True
        )
        # mask (1= model this pixel, 0= leave blanck)

        kwargs_data = {
            "ra_at_xy_0": ra_at_xy_0,
            "dec_at_xy_0": dec_at_xy_0,
            "transform_pix2angle": Mpix2coord,
            "image_data": np.ones((numPix, numPix)),
        }
        data = ImageData(**kwargs_data)

        ra_shift = 0.05
        dec_shift = 0.0
        kwargs_data["ra_shift"] = ra_shift
        kwargs_data["dec_shift"] = dec_shift
        data_shift = ImageData(**kwargs_data)

        ra, dec = data.map_pix2coord(1, 1)
        ra_new, dec_new = data_shift.map_pix2coord(1, 1)
        npt.assert_almost_equal(ra_new - ra, ra_shift, decimal=10)
        npt.assert_almost_equal(dec_new - dec, dec_shift, decimal=10)

        ra_2, dec_2 = data_shift.map_pix2coord(2, 1)
        npt.assert_almost_equal(ra, ra_2, decimal=10)
        npt.assert_almost_equal(dec, dec_2, decimal=10)

        x, y = data.map_coord2pix(0, 0)
        x_new, y_new = data_shift.map_coord2pix(ra_shift, dec_shift)
        npt.assert_almost_equal(x, x_new, decimal=10)
        npt.assert_almost_equal(y, y_new, decimal=10)

    def test_shift_coordinate_system(self):
        x_shift = 0.05
        y_shift = 0

        numPix = 10
        deltaPix = 0.05
        (
            x_grid,
            y_grid,
            ra_at_xy_0,
            dec_at_xy_0,
            x_at_radec_0,
            y_at_radec_0,
            Mpix2coord,
            Mcoord2pix,
        ) = util.make_grid_with_coordtransform(
            numPix=numPix, deltapix=deltaPix, subgrid_res=1, inverse=True
        )

        kwargs_data = {
            "ra_at_xy_0": ra_at_xy_0,
            "dec_at_xy_0": dec_at_xy_0,
            "transform_pix2angle": Mpix2coord,
            "image_data": np.ones((numPix, numPix)),
        }

        data = ImageData(**kwargs_data)
        data_new = copy.deepcopy(data)
        data_new.shift_coordinate_system(x_shift, y_shift, pixel_unit=False)
        ra, dec = 0, 0
        x, y = data.map_coord2pix(ra, dec)
        x_new, y_new = data_new.map_coord2pix(ra + x_shift, dec + y_shift)
        npt.assert_almost_equal(x, x_new, decimal=10)
        npt.assert_almost_equal(y, y_new, decimal=10)

        ra, dec = data.map_pix2coord(x, y)
        ra_new, dec_new = data_new.map_pix2coord(x, y)
        npt.assert_almost_equal(ra, ra_new - x_shift, decimal=10)
        npt.assert_almost_equal(dec, dec_new - y_shift, decimal=10)

        x_coords, y_coords = data.pixel_coordinates
        x_coords_new, y_coords_new = data_new.pixel_coordinates
        npt.assert_almost_equal(x_coords[0], x_coords_new[0] - x_shift, decimal=10)
        npt.assert_almost_equal(y_coords[0], y_coords_new[0] - y_shift, decimal=10)

    def test_update_data(self):
        kwargs_data = {
            "image_data": np.zeros((self.numPix, self.numPix)),
            "noise_map": None,
            "exposure_time": 1,
            "background_rms": 1,
        }
        data = ImageData(**kwargs_data)
        C_D = data.C_D
        data.update_data(image_data=np.ones((self.numPix, self.numPix)))
        C_D_new = data.C_D
        assert C_D_new[0, 0] > C_D[0, 0]
        data_new = data.data
        npt.assert_almost_equal(data_new, np.ones((self.numPix, self.numPix)))


class TestRaise(unittest.TestCase):
    def test_raise(self):
        kwargs_data = {"image_data": np.zeros((10, 10))}
        Data = ImageData(**kwargs_data)
        image_data_new = np.zeros((5, 5))
        with self.assertRaises(ValueError):
            out = Data.update_data(image_data_new)
        with self.assertRaises(ValueError):
            ImageData(**kwargs_data, likelihood_method="WRONG")


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

#  this file contains a class to make a Sersic profile

import numpy as np
from lenstronomy.LensModel.Profiles.sersic_utils import SersicUtil
import lenstronomy.Util.param_util as param_util
import numpy as np
from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
class Sersic(SersicUtil):
    """This class contains functions to evaluate a spherical Sersic function.

    .. math::
        I(R) = I_0 \\exp \\left[ -b_n (R/R_{\\rm Sersic})^{\\frac{1}{n}}\\right]

    with :math:`I_0 = amp`
    and
    with :math:`b_{n}\\approx 1.999n-0.327`
    """

    param_names = ["amp", "R_sersic", "n_sersic", "center_x", "center_y"]
    lower_limit_default = {
        "amp": 0,
        "R_sersic": 0,
        "n_sersic": 0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "R_sersic": 100,
        "n_sersic": 8,
        "center_x": 100,
        "center_y": 100,
    }

    def function(
        self, x, y, amp, R_sersic, n_sersic, center_x=0, center_y=0, max_R_frac=1000.0
    ):
        """

        :param x:
        :param y:
        :param amp: surface brightness/amplitude value at the half light radius
        :param R_sersic: semi-major axis half light radius
        :param n_sersic: Sersic index
        :param center_x: center in x-coordinate
        :param center_y: center in y-coordinate
        :param max_R_frac: maximum window outside which the mass is zeroed, in units of R_sersic (float)
        :return: Sersic profile value at (x, y)
        """
        R = self.get_distance_from_center(
            x, y, e1=0, e2=0, center_x=center_x, center_y=center_y
        )
        result = self._r_sersic(R, R_sersic, n_sersic, max_R_frac)
        return amp * result


@export
class SersicElliptic(SersicUtil):
    """This class contains functions to evaluate an elliptical Sersic function.

    .. math::

        I(R) = I_0 \\exp \\left[ -b_n (R/R_{\\rm Sersic})^{\\frac{1}{n}}\\right]

    with :math:`I_0 = amp`,
    :math:`R = \\sqrt{q \\theta^2_x + \\theta^2_y/q}`
    and
    with :math:`b_{n}\\approx 1.999n-0.327`
    """

    param_names = ["amp", "R_sersic", "n_sersic", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "amp": 0,
        "R_sersic": 0,
        "n_sersic": 0.5,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "R_sersic": 100,
        "n_sersic": 8,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def function(
        self,
        x,
        y,
        amp,
        R_sersic,
        n_sersic,
        e1,
        e2,
        center_x=0,
        center_y=0,
        max_R_frac=1000.0,
    ):
        """

        :param x:
        :param y:
        :param amp: surface brightness/amplitude value at the half light radius
        :param R_sersic: half light radius (either semi-major axis or product average of semi-major and semi-minor axis)
        :param n_sersic: Sersic index
        :param e1: eccentricity parameter e1
        :param e2: eccentricity parameter e2
        :param center_x: center in x-coordinate
        :param center_y: center in y-coordinate
        :param max_R_frac: maximum window outside which the mass is zeroed, in units of R_sersic (float)
        :return: Sersic profile value at (x, y)
        """

        R_sersic = np.maximum(0, R_sersic)
        R = self.get_distance_from_center(x, y, e1, e2, center_x, center_y)
        result = self._r_sersic(R, R_sersic, n_sersic, max_R_frac)
        return amp * result


@export
class SersicElliptic_qPhi(SersicUtil):
    """This class is the same as SersicElliptic except sampling over q and phi instead
    of e1 and e2."""

    param_names = ["amp", "R_sersic", "n_sersic", "q", "phi", "center_x", "center_y"]
    lower_limit_default = {
        "amp": 0,
        "R_sersic": 0,
        "n_sersic": 0.5,
        "q": 0,
        "phi": -np.pi,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "R_sersic": 100,
        "n_sersic": 8,
        "q": 1.0,
        "phi": np.pi,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self, *args, **kwargs):
        self._sersic_e1e2 = SersicElliptic(*args, **kwargs)

    def function(
        self,
        x,
        y,
        amp,
        R_sersic,
        n_sersic,
        q,
        phi,
        center_x=0,
        center_y=0,
        max_R_frac=100.0,
    ):
        """

        :param x:
        :param y:
        :param amp: surface brightness/amplitude value at the half light radius
        :param R_sersic: half light radius (either semi-major axis or product average of semi-major and semi-minor axis)
        :param n_sersic: Sersic index
        :param q: axis ratio
        :param phi: position angle (radians)
        :param center_x: center in x-coordinate
        :param center_y: center in y-coordinate
        :param max_R_frac: maximum window outside of which the mass is zeroed, in units of R_sersic (float)
        :return: Sersic profile value at (x, y)
        """

        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        return self._sersic_e1e2.function(
            x, y, amp, R_sersic, n_sersic, e1, e2, center_x, center_y, max_R_frac
        )


@export
class CoreSersic(SersicUtil):
    """This class contains the Core-Sersic function introduced by e.g. Trujillo et al.
    2004.

    .. math::

        I(R) = I' \\left[1 + (R_b/R)^{\\alpha} \\right]^{\\gamma / \\alpha}
        \\exp \\left{ -b_n \\left[(R^{\\alpha} + R_b^{\\alpha})/R_e^{\\alpha}  \\right]^{1 / (n\\alpha)}  \\right}

    with

    .. math::
        I' = I_b 2^{-\\gamma/ \\alpha} \\exp \\left[b_n 2^{1 / (n\\alpha)} (R_b/R_e)^{1/n}  \\right]

    where :math:`I_b` is the intensity at the break radius and :math:`R = \\sqrt{q \\theta^2_x + \\theta^2_y/q}`.
    """

    param_names = [
        "amp",
        "R_sersic",
        "Rb",
        "n_sersic",
        "gamma",
        "e1",
        "e2",
        "center_x",
        "center_y",
    ]
    lower_limit_default = {
        "amp": 0,
        "Rb": 0,
        "n_sersic": 0.5,
        "gamma": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "Rb": 100,
        "n_sersic": 8,
        "gamma": 10,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def function(
        self,
        x,
        y,
        amp,
        R_sersic,
        Rb,
        n_sersic,
        gamma,
        e1,
        e2,
        center_x=0,
        center_y=0,
        alpha=3.0,
        max_R_frac=1000.0,
    ):
        """
        :param x:
        :param y:
        :param amp: surface brightness/amplitude value at the half light radius
        :param R_sersic: half light radius (either semi-major axis or product average of semi-major and semi-minor axis)
        :param Rb: "break" core radius
        :param n_sersic: Sersic index
        :param gamma: inner power-law exponent
        :param e1: eccentricity parameter e1
        :param e2: eccentricity parameter e2
        :param center_x: center in x-coordinate
        :param center_y: center in y-coordinate
        :param alpha: sharpness of the transition between the cusp and the outer Sersic profile (float)
        :param max_R_frac: maximum window outside which the mass is zeroed, in units of R_sersic (float)
        :return: Cored Sersic profile value at (x, y)
        """
        # TODO: max_R_frac not implemented
        R_ = self.get_distance_from_center(x, y, e1, e2, center_x, center_y)
        R = self._R_stable(R_)
        bn = self.b_n(n_sersic)
        result = (
            amp
            * (1 + (Rb / R) ** alpha) ** (gamma / alpha)
            * np.exp(
                -bn
                * (
                    ((R**alpha + Rb**alpha) / R_sersic**alpha)
                    ** (1.0 / (alpha * n_sersic))
                    - 1.0
                )
            )
        )
        return np.nan_to_num(result)



__author__ = "sibirrer"
# this file contains a class to make a gaussian

import numpy as np
from lenstronomy.LensModel.Profiles.sersic import Sersic
import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["SersicEllipse"]


class SersicEllipse(LensProfileBase):
    """
    this class contains functions to evaluate a Sersic mass profile: https://arxiv.org/pdf/astro-ph/0311559.pdf
    """

    param_names = ["k_eff", "R_sersic", "n_sersic", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "k_eff": 0,
        "R_sersic": 0,
        "n_sersic": 0.5,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "k_eff": 10,
        "R_sersic": 100,
        "n_sersic": 8,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self.sersic = Sersic()
        self._diff = 0.000001
        super(SersicEllipse, self).__init__()

    def function(self, x, y, n_sersic, R_sersic, k_eff, e1, e2, center_x=0, center_y=0):
        """Returns Gaussian."""
        # phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        x_, y_ = param_util.transform_e1e2_square_average(
            x, y, e1, e2, center_x, center_y
        )
        # x_, y_ = self._coord_transf(x, y, q, phi_G, center_x, center_y)
        f_ = self.sersic.function(x_, y_, n_sersic, R_sersic, k_eff)
        return f_

    def derivatives(
        self, x, y, n_sersic, R_sersic, k_eff, e1, e2, center_x=0, center_y=0
    ):
        """Returns df/dx and df/dy of the function."""
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        e = param_util.q2e(q)
        # e = abs(1. - q)
        cos_phi = np.cos(phi_G)
        sin_phi = np.sin(phi_G)
        x_, y_ = param_util.transform_e1e2_square_average(
            x, y, e1, e2, center_x, center_y
        )
        # x_, y_ = self._coord_transf(x, y, q, phi_G, center_x, center_y)
        f_x_prim, f_y_prim = self.sersic.derivatives(x_, y_, n_sersic, R_sersic, k_eff)
        f_x_prim *= np.sqrt(1 - e)
        f_y_prim *= np.sqrt(1 + e)
        f_x = cos_phi * f_x_prim - sin_phi * f_y_prim
        f_y = sin_phi * f_x_prim + cos_phi * f_y_prim
        return f_x, f_y

    def hessian(self, x, y, n_sersic, R_sersic, k_eff, e1, e2, center_x=0, center_y=0):
        """Returns Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx,
        d^f/dy^2."""
        alpha_ra, alpha_dec = self.derivatives(
            x, y, n_sersic, R_sersic, k_eff, e1, e2, center_x, center_y
        )
        diff = self._diff
        alpha_ra_dx, alpha_dec_dx = self.derivatives(
            x + diff, y, n_sersic, R_sersic, k_eff, e1, e2, center_x, center_y
        )
        alpha_ra_dy, alpha_dec_dy = self.derivatives(
            x, y + diff, n_sersic, R_sersic, k_eff, e1, e2, center_x, center_y
        )

        f_xx = (alpha_ra_dx - alpha_ra) / diff
        f_xy = (alpha_ra_dy - alpha_ra) / diff
        f_yx = (alpha_dec_dx - alpha_dec) / diff
        f_yy = (alpha_dec_dy - alpha_dec) / diff

        return f_xx, f_xy, f_yx, f_yy



__author__ = ["nataliehogg", "pierrefleury"]

from lenstronomy.LensModel.LineOfSight.LOSModels.los import LOS

__all__ = ["LOSMinimal"]


class LOSMinimal(LOS):
    """
    Class deriving from LOS containing the parameters for line-of-sight
    corrections within the "minimal model" defined in
    https://arxiv.org/abs/2104.08883
    It is equivalent to LOS but with fewer parameters, namely:
    kappa_od, gamma1_od, gamma2_od, omega_od, kappa_los, gamma1_los,
    gamma2_los, omega_los.
    """

    param_names = [
        "kappa_od",
        "gamma1_od",
        "gamma2_od",
        "omega_od",
        "kappa_los",
        "gamma1_los",
        "gamma2_los",
        "omega_los",
    ]
    lower_limit_default = {pert: -0.5 for pert in param_names}
    upper_limit_default = {pert: 0.5 for pert in param_names}



__author__ = "sibirrer"

# this file contains a class which describes the surface brightness of the light models

import numpy as np
from lenstronomy.LightModel.light_model_base import LightModelBase

__all__ = ["LinearBasis"]


class LinearBasis(LightModelBase):
    """Class to handle source and lens light models."""

    def __init__(self, **kwargs):
        """

        :param kwargs: keyword arguments for LightModelBase class
        """
        super(LinearBasis, self).__init__(**kwargs)

    @property
    def param_name_list(self):
        """Returns the list of all parameter names.

        :return: list of list of strings (for each light model separately)
        """
        name_list = []
        for i, func in enumerate(self.func_list):
            name_list.append(func.param_names)
        return name_list

    @property
    def param_name_list_latex(self):
        """Returns the list of all parameter names in LateX style.

        :return: list of list of strings (for each light model separately)
        """
        name_list = []
        for i, func in enumerate(self.func_list):
            if hasattr(func, "param_names_latex"):
                name_list.append(func.param_names_latex)
            else:
                name_list.append(func.param_names)
        return name_list

    def functions_split(self, x, y, kwargs_list, k=None):
        """Split model in different components.

        :param x: coordinate in units of arcsec relative to the center of the image
        :param y: coordinate in units of arcsec relative to the center of the image
        :param kwargs_list: keyword argument list of light profile
        :param k: integer or list of integers for selecting subsets of light profiles
        """
        response = []
        n = 0
        bool_list = self._bool_list(k=k)
        for i, model in enumerate(self.profile_type_list):
            if bool_list[i] is True:
                if model in [
                    "SERSIC",
                    "SERSIC_ELLIPSE",
                    "SERSIC_ELLIPSE_Q_PHI",
                    "CORE_SERSIC",
                    "HERNQUIST",
                    "HERNQUIST_ELLIPSE",
                    "PJAFFE",
                    "PJAFFE_ELLIPSE",
                    "GAUSSIAN",
                    "GAUSSIAN_ELLIPSE",
                    "POWER_LAW",
                    "NIE",
                    "CHAMELEON",
                    "DOUBLE_CHAMELEON",
                    "TRIPLE_CHAMELEON",
                    "UNIFORM",
                    "INTERPOL",
                    "ELLIPSOID",
                ]:
                    kwargs_new = kwargs_list[i].copy()
                    new = {"amp": 1}
                    kwargs_new.update(new)
                    response += [self.func_list[i].function(x, y, **kwargs_new)]
                    n += 1
                elif model in ["MULTI_GAUSSIAN", "MULTI_GAUSSIAN_ELLIPSE"]:
                    num = len(kwargs_list[i]["amp"])
                    new = {"amp": np.ones(num)}
                    kwargs_new = kwargs_list[i].copy()
                    kwargs_new.update(new)
                    response += self.func_list[i].function_split(x, y, **kwargs_new)
                    n += num
                elif model in [
                    "SHAPELETS",
                    "SHAPELETS_POLAR",
                    "SHAPELETS_POLAR_EXP",
                    "SHAPELETS_ELLIPSE",
                ]:
                    kwargs = kwargs_list[i]
                    n_max = kwargs["n_max"]
                    if model in ["SHAPELETS_POLAR_EXP"]:
                        num_param = int((n_max + 1) ** 2)
                    else:
                        num_param = int((n_max + 1) * (n_max + 2) / 2)
                    new = {"amp": np.ones(num_param)}
                    kwargs_new = kwargs_list[i].copy()
                    kwargs_new.update(new)
                    response += self.func_list[i].function_split(x, y, **kwargs_new)
                    n += num_param
                elif model in ["SLIT_STARLETS", "SLIT_STARLETS_GEN2"]:
                    raise ValueError(
                        "'{}' model does not support function split".format(model)
                    )
                else:
                    raise ValueError("model type %s not valid!" % model)
        return response, n

    def num_param_linear(self, kwargs_list, list_return=False):
        """

        :param kwargs_list: list of keyword arguments of the light profiles
        :param list_return: bool, if True returns list of individual number of parameters
        :return: number of linear basis set coefficients
        """
        n_list = self.num_param_linear_list(kwargs_list)
        if not list_return:
            return np.sum(n_list)
        return n_list

    def num_param_linear_list(self, kwargs_list):
        """Returns the list (in order of the light profiles) of the number of linear
        components per model.

        :param kwargs_list: list of keyword arguments of the light profiles
        :return: number of linear basis set coefficients
        """
        n_list = []
        for i, model in enumerate(self.profile_type_list):
            if model in [
                "SERSIC",
                "SERSIC_ELLIPSE",
                "CORE_SERSIC",
                "HERNQUIST",
                "HERNQUIST_ELLIPSE",
                "PJAFFE",
                "PJAFFE_ELLIPSE",
                "GAUSSIAN",
                "GAUSSIAN_ELLIPSE",
                "POWER_LAW",
                "NIE",
                "CHAMELEON",
                "DOUBLE_CHAMELEON",
                "TRIPLE_CHAMELEON",
                "UNIFORM",
                "INTERPOL",
                "ELLIPSOID",
            ]:
                n_list += [1]
            elif model in ["MULTI_GAUSSIAN", "MULTI_GAUSSIAN_ELLIPSE"]:
                num = len(kwargs_list[i]["sigma"])
                n_list += [num]
            elif model in [
                "SHAPELETS",
                "SHAPELETS_POLAR",
                "SHAPELETS_POLAR_EXP",
                "SHAPELETS_ELLIPSE",
            ]:
                n_max = kwargs_list[i]["n_max"]
                if model in ["SHAPELETS_POLAR_EXP"]:
                    num_param = int((n_max + 1) ** 2)
                else:
                    num_param = int((n_max + 1) * (n_max + 2) / 2)
                n_list += [num_param]
            elif model in ["SLIT_STARLETS", "SLIT_STARLETS_GEN2"]:
                n_scales = kwargs_list[i]["n_scales"]
                n_pixels = kwargs_list[i]["n_pixels"]
                num_param = int(n_scales * n_pixels)
                n_list += [
                    num_param
                ]  # TODO : find a way to make it the number of source pixels
            else:
                raise ValueError("model type %s not valid!" % model)
        return n_list

    def update_linear(self, param, i, kwargs_list):
        """

        :param param: array of linear amplitude coefficients in the order of the linear minimization of the ImSim module
        :param i: index of first coefficient to start reading out the linear parameters associated with the model
         components of this class
        :param kwargs_list: list of keyword arguments of the model components
        :return: kwargs list with over-written or added 'amp' parameters according to the coefficients in param
        """
        for k, model in enumerate(self.profile_type_list):
            if model in [
                "SERSIC",
                "SERSIC_ELLIPSE",
                "SERSIC_ELLIPSE_Q_PHI",
                "CORE_SERSIC",
                "HERNQUIST",
                "PJAFFE",
                "PJAFFE_ELLIPSE",
                "HERNQUIST_ELLIPSE",
                "GAUSSIAN",
                "GAUSSIAN_ELLIPSE",
                "POWER_LAW",
                "NIE",
                "CHAMELEON",
                "DOUBLE_CHAMELEON",
                "TRIPLE_CHAMELEON",
                "UNIFORM",
                "INTERPOL",
                "ELLIPSOID",
            ]:
                kwargs_list[k]["amp"] = param[i]
                i += 1
            elif model in ["MULTI_GAUSSIAN", "MULTI_GAUSSIAN_ELLIPSE"]:
                num_param = len(kwargs_list[k]["sigma"])
                kwargs_list[k]["amp"] = param[i : i + num_param]
                i += num_param
            elif model in [
                "SHAPELETS",
                "SHAPELETS_POLAR",
                "SHAPELETS_POLAR_EXP",
                "SHAPELETS_ELLIPSE",
            ]:
                n_max = kwargs_list[k]["n_max"]
                if model in ["SHAPELETS_POLAR_EXP"]:
                    num_param = int((n_max + 1) ** 2)
                else:
                    num_param = int((n_max + 1) * (n_max + 2) / 2)
                kwargs_list[k]["amp"] = param[i : i + num_param]
                i += num_param
            elif model in ["SLIT_STARLETS", "SLIT_STARLETS_GEN2"]:
                n_scales = kwargs_list[k]["n_scales"]
                n_pixels = kwargs_list[k]["n_pixels"]
                num_param = int(n_scales * n_pixels)
                kwargs_list[k]["amp"] = param[i : i + num_param]
                i += num_param
            else:
                raise ValueError("model type %s not valid!" % model)
        return kwargs_list, i

    def add_fixed_linear(self, kwargs_fixed_list):
        """

        :param kwargs_fixed_list: list of fixed keyword arguments
        :return: updated kwargs_fixed_list with additional linear parameters being fixed.
        """
        for k, model in enumerate(self.profile_type_list):
            kwargs_fixed = kwargs_fixed_list[k]
            param_names = self.param_name_list[k]
            if "amp" in param_names:
                if "amp" not in kwargs_fixed:
                    kwargs_fixed["amp"] = 1
        return kwargs_fixed_list

    def linear_param_from_kwargs(self, kwargs_list):
        """Inverse function of update_linear() returning the linear amplitude list for
        the keyword argument list.

        :param kwargs_list: model parameters including the linear amplitude parameters
        :type kwargs_list: list of keyword arguments
        :return: list of linear amplitude parameters
        :rtype: list
        """
        param = []
        for k, model in enumerate(self.profile_type_list):
            kwargs_ = kwargs_list[k]
            param_names = self.param_name_list[k]
            if "amp" in param_names:
                amp = kwargs_["amp"]
                amp_list = np.atleast_1d(amp)
                for a in amp_list:
                    param.append(a)
        return param

    def check_positive_flux_profile(self, kwargs_list):
        """Check whether linear amplitude parameter are non-negative for specified list
        of lens models that have a physical amplitude interpretation.

        :param kwargs_list: light model parameter keyword argument list
        :return: bool, if True, no specified model has negative flux
        """
        pos_bool = True
        for k, model in enumerate(self.profile_type_list):
            if "amp" in kwargs_list[k]:
                if model in [
                    "SERSIC",
                    "SERSIC_ELLIPSE",
                    "CORE_SERSIC",
                    "HERNQUIST",
                    "PJAFFE",
                    "PJAFFE_ELLIPSE",
                    "HERNQUIST_ELLIPSE",
                    "GAUSSIAN",
                    "GAUSSIAN_ELLIPSE",
                    "POWER_LAW",
                    "NIE",
                    "CHAMELEON",
                    "DOUBLE_CHAMELEON",
                ]:
                    if kwargs_list[k]["amp"] < 0:
                        pos_bool = False
                        break
        return pos_bool



import pytest
import numpy as np
from lenstronomy.Util.param_util import ellipticity2phi_q
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.QuadOptimizer.param_manager import (
    PowerLawFixedShear,
    PowerLawFixedShearMultipole,
    PowerLawFreeShear,
    PowerLawFreeShearMultipole,
)
from lenstronomy.LensModel.QuadOptimizer.optimizer import Optimizer
import numpy.testing as npt


class TestOptimizer(object):
    def setup_method(self):
        self.zlens, self.zsource = 0.5, 1.5
        epl_kwargs = {
            "theta_E": 1.0,
            "center_x": 0.0,
            "center_y": 0.0,
            "e1": 0.2,
            "e2": 0.1,
            "gamma": 2.05,
        }
        shear_kwargs = {"gamma1": 0.05, "gamma2": -0.04}
        kwargs_macro = [epl_kwargs, shear_kwargs]

        self.x_image = np.array([0.65043538, -0.31109505, 0.78906059, -0.86222271])
        self.y_image = np.array([-0.89067493, 0.94851787, 0.52882605, -0.25403778])

        halo_list = ["SIS", "SIS", "SIS"]
        halo_z = [self.zlens - 0.1, self.zlens, self.zlens + 0.4]
        halo_kwargs = [
            {"theta_E": 0.1, "center_x": 0.3, "center_y": -0.9},
            {"theta_E": 0.15, "center_x": 1.3, "center_y": -0.5},
            {"theta_E": 0.06, "center_x": -0.4, "center_y": -0.4},
        ]

        self.kwargs_epl = kwargs_macro + halo_kwargs
        self.zlist_epl = [self.zlens, self.zlens] + halo_z
        self.lens_model_list_epl = ["EPL", "SHEAR"] + halo_list

        kwargs_multi = [
            {"m": 4, "a_m": -0.04, "phi_m": -0.2, "center_x": 0.1, "center_y": -0.1}
        ]
        self.kwargs_multipole = kwargs_macro + kwargs_multi + halo_kwargs
        self.zlist_multipole = [self.zlens, self.zlens, self.zlens] + halo_z
        self.lens_model_list_multipole = ["EPL", "SHEAR"] + ["MULTIPOLE"] + halo_list

    def test_elp_free_shear(self):
        param_class = PowerLawFreeShear(self.kwargs_epl)

        optimizer = Optimizer(
            self.x_image,
            self.y_image,
            self.lens_model_list_epl,
            self.zlist_epl,
            self.zlens,
            self.zsource,
            param_class,
            pso_convergence_mean=50000,
            foreground_rays=None,
            tol_source=1e-5,
            tol_simplex_func=1e-3,
            simplex_n_iterations=400,
        )

        kwargs_final, source = optimizer.optimize(50, 100, verbose=True)

        lensmodel = LensModel(
            self.lens_model_list_epl,
            self.zlens,
            self.zsource,
            self.zlist_epl,
            multi_plane=True,
        )
        beta_x, beta_y = lensmodel.ray_shooting(
            self.x_image, self.y_image, kwargs_final
        )

        npt.assert_almost_equal(np.sum(beta_x) - 4 * np.mean(beta_x), 0)
        npt.assert_almost_equal(np.sum(beta_y) - 4 * np.mean(beta_y), 0)

    def test_elp_fixed_shear(self):
        param_class = PowerLawFixedShear(self.kwargs_epl, 0.06)

        optimizer = Optimizer(
            self.x_image,
            self.y_image,
            self.lens_model_list_epl,
            self.zlist_epl,
            self.zlens,
            self.zsource,
            param_class,
            pso_convergence_mean=50000,
            foreground_rays=None,
            tol_source=1e-5,
            tol_simplex_func=1e-3,
            simplex_n_iterations=400,
        )

        kwargs_final, source = optimizer.optimize(50, 100, verbose=True)

        lensmodel = LensModel(
            self.lens_model_list_epl,
            self.zlens,
            self.zsource,
            self.zlist_epl,
            multi_plane=True,
        )
        beta_x, beta_y = lensmodel.ray_shooting(
            self.x_image, self.y_image, kwargs_final
        )

        npt.assert_almost_equal(np.sum(beta_x) - 4 * np.mean(beta_x), 0)
        npt.assert_almost_equal(np.sum(beta_y) - 4 * np.mean(beta_y), 0)

        kwargs_shear = kwargs_final[1]
        shear_out = np.hypot(kwargs_shear["gamma1"], kwargs_shear["gamma2"])
        npt.assert_almost_equal(shear_out, 0.06)

    def test_multipole_free_shear(self):
        param_class = PowerLawFreeShearMultipole(self.kwargs_multipole)

        optimizer = Optimizer(
            self.x_image,
            self.y_image,
            self.lens_model_list_multipole,
            self.zlist_multipole,
            self.zlens,
            self.zsource,
            param_class,
            pso_convergence_mean=50000,
            foreground_rays=None,
            tol_source=1e-5,
            tol_simplex_func=1e-3,
            simplex_n_iterations=400,
        )

        kwargs_final, source = optimizer.optimize(50, 100, verbose=True)

        lensmodel = LensModel(
            self.lens_model_list_multipole,
            self.zlens,
            self.zsource,
            self.zlist_multipole,
            multi_plane=True,
        )
        beta_x, beta_y = lensmodel.ray_shooting(
            self.x_image, self.y_image, kwargs_final
        )

        npt.assert_almost_equal(np.sum(beta_x) - 4 * np.mean(beta_x), 0)
        npt.assert_almost_equal(np.sum(beta_y) - 4 * np.mean(beta_y), 0)

        kwargs_epl = kwargs_final[0]
        kwargs_multipole = kwargs_final[2]
        npt.assert_almost_equal(kwargs_multipole["m"], 4)
        npt.assert_almost_equal(kwargs_multipole["center_x"], kwargs_epl["center_x"])
        npt.assert_almost_equal(kwargs_multipole["center_y"], kwargs_epl["center_y"])
        phi, _ = ellipticity2phi_q(kwargs_epl["e1"], kwargs_epl["e2"])
        npt.assert_almost_equal(phi, kwargs_multipole["phi_m"])

    def test_multipole_fixed_shear(self):
        param_class = PowerLawFixedShearMultipole(self.kwargs_multipole, 0.07)

        optimizer = Optimizer(
            self.x_image,
            self.y_image,
            self.lens_model_list_multipole,
            self.zlist_multipole,
            self.zlens,
            self.zsource,
            param_class,
            pso_convergence_mean=50000,
            foreground_rays=None,
            tol_source=1e-5,
            tol_simplex_func=1e-3,
            simplex_n_iterations=400,
        )

        kwargs_final, source = optimizer.optimize(50, 100, verbose=True)

        lensmodel = LensModel(
            self.lens_model_list_multipole,
            self.zlens,
            self.zsource,
            self.zlist_multipole,
            multi_plane=True,
        )
        beta_x, beta_y = lensmodel.ray_shooting(
            self.x_image, self.y_image, kwargs_final
        )

        npt.assert_almost_equal(np.sum(beta_x) - 4 * np.mean(beta_x), 0)
        npt.assert_almost_equal(np.sum(beta_y) - 4 * np.mean(beta_y), 0)

        kwargs_shear = kwargs_final[1]
        shear_out = np.hypot(kwargs_shear["gamma1"], kwargs_shear["gamma2"])
        npt.assert_almost_equal(shear_out, 0.07)

        kwargs_epl = kwargs_final[0]
        kwargs_multipole = kwargs_final[2]
        npt.assert_almost_equal(kwargs_multipole["m"], 4)
        npt.assert_almost_equal(kwargs_multipole["center_x"], kwargs_epl["center_x"])
        npt.assert_almost_equal(kwargs_multipole["center_y"], kwargs_epl["center_y"])
        phi, _ = ellipticity2phi_q(kwargs_epl["e1"], kwargs_epl["e2"])
        npt.assert_almost_equal(phi, kwargs_multipole["phi_m"])

    def test_options(self):
        param_class = PowerLawFixedShearMultipole(self.kwargs_multipole, 0.07)

        optimizer = Optimizer(
            self.x_image,
            self.y_image,
            self.lens_model_list_multipole,
            self.zlist_multipole,
            self.zlens,
            self.zsource,
            param_class,
            pso_convergence_mean=50000,
            particle_swarm=False,
            foreground_rays=None,
            tol_source=1e-5,
            tol_simplex_func=1e-3,
            simplex_n_iterations=400,
        )

        kwargs_final, source = optimizer.optimize(50, 100, verbose=True)

        lensmodel = LensModel(
            self.lens_model_list_multipole,
            self.zlens,
            self.zsource,
            self.zlist_multipole,
            multi_plane=True,
        )
        beta_x, beta_y = lensmodel.ray_shooting(
            self.x_image, self.y_image, kwargs_final
        )

        npt.assert_almost_equal(np.sum(beta_x) - 4 * np.mean(beta_x), 0)
        npt.assert_almost_equal(np.sum(beta_y) - 4 * np.mean(beta_y), 0)

        kwargs_shear = kwargs_final[1]
        shear_out = np.hypot(kwargs_shear["gamma1"], kwargs_shear["gamma2"])
        npt.assert_almost_equal(shear_out, 0.07)

        foreground_rays = optimizer.fast_rayshooting._foreground_rays
        optimizer = Optimizer(
            self.x_image,
            self.y_image,
            self.lens_model_list_multipole,
            self.zlist_multipole,
            self.zlens,
            self.zsource,
            param_class,
            pso_convergence_mean=50000,
            particle_swarm=False,
            re_optimize=True,
            re_optimize_scale=0.5,
            foreground_rays=foreground_rays,
            tol_source=1e-5,
            tol_simplex_func=1e-3,
            simplex_n_iterations=400,
        )

        kwargs_final, source = optimizer.optimize(50, 100, verbose=True)

        lensmodel = LensModel(
            self.lens_model_list_multipole,
            self.zlens,
            self.zsource,
            self.zlist_multipole,
            multi_plane=True,
        )
        beta_x, beta_y = lensmodel.ray_shooting(
            self.x_image, self.y_image, kwargs_final
        )

        npt.assert_almost_equal(np.sum(beta_x) - 4 * np.mean(beta_x), 0)
        npt.assert_almost_equal(np.sum(beta_y) - 4 * np.mean(beta_y), 0)

        kwargs_shear = kwargs_final[1]
        shear_out = np.hypot(kwargs_shear["gamma1"], kwargs_shear["gamma2"])
        npt.assert_almost_equal(shear_out, 0.07)

    def test_multi_threading(self):
        param_class = PowerLawFixedShearMultipole(self.kwargs_multipole, 0.07)

        optimizer = Optimizer(
            self.x_image,
            self.y_image,
            self.lens_model_list_multipole,
            self.zlist_multipole,
            self.zlens,
            self.zsource,
            param_class,
            pso_convergence_mean=50000,
            particle_swarm=True,
            foreground_rays=None,
            tol_source=1e-5,
            tol_simplex_func=1e-3,
            simplex_n_iterations=400,
        )

        kwargs_final, source = optimizer.optimize(50, 100, verbose=True, threadCount=5)
        lensmodel = LensModel(
            self.lens_model_list_multipole,
            self.zlens,
            self.zsource,
            self.zlist_multipole,
            multi_plane=True,
        )
        beta_x, beta_y = lensmodel.ray_shooting(
            self.x_image, self.y_image, kwargs_final
        )

        npt.assert_almost_equal(np.sum(beta_x) - 4 * np.mean(beta_x), 0)
        npt.assert_almost_equal(np.sum(beta_y) - 4 * np.mean(beta_y), 0)


if __name__ == "__main__":
    pytest.main()



import pytest
from lenstronomy.LightModel.Profiles.uniform import Uniform


class TestShapelet(object):
    """Class to test Shapelets."""

    def setup_method(self):
        pass

    def test_function(self):
        """

        :return:
        """
        uniform = Uniform()
        output = uniform.function(x=1, y=1, amp=0.1)
        assert output == 0.1


if __name__ == "__main__":
    pytest.main()



import numpy as np
import lenstronomy.Util.param_util as param_util
from lenstronomy.LightModel.Profiles.profile_base import LightProfileBase

__all__ = ["NIE"]


class NIE(LightProfileBase):
    """Non-divergent isothermal ellipse (projected) This is effectively the convergence
    profile of the NIE lens model with an amplitude 'amp' rather than an Einstein radius
    'theta_E'."""

    param_names = ["amp", "e1", "e2", "s_scale", "center_x", "center_y"]
    lower_limit_default = {
        "amp": 0,
        "e1": -0.5,
        "e2": -0.5,
        "s_scale": 0,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "e1": 0.5,
        "e2": 0.5,
        "s_scale": 100,
        "center_x": 100,
        "center_y": 100,
    }

    def function(self, x, y, amp, e1, e2, s_scale, center_x=0, center_y=0):
        """

        :param x: x-coordinate
        :param y: y-coordinate
        :param amp: surface brightness normalization
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param s_scale: smoothing scale (square averaged of minor and major axis)
        :param center_x: center of profile
        :param center_y: center of profile
        :return: surface brightness of NIE profile
        """
        # phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        # s = s_scale * np.sqrt((1 + q ** 2) / (2 * q ** 2))
        x__, y__ = param_util.transform_e1e2_product_average(
            x, y, e1, e2, center_x, center_y
        )
        f_ = amp / 2.0 * (s_scale**2 + x__**2 + y__**2) ** (-1.0 / 2)
        return f_

    def light_3d(self, r, amp, e1, e2, s_scale, center_x=0, center_y=0):
        """3d light distribution (in spherical regime)

        :param r: 3d radius
        :param amp: surface brightness normalization
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param s_scale: smoothing scale (square averaged of minor and major axis)
        :param center_x: center of profile
        :param center_y: center of profile
        :return: light density at 3d radius
        """
        rho = self._amp2rho(amp)
        return rho / (r**2 + s_scale**2)

    @staticmethod
    def _amp2rho(amp):
        """Converts surface brightness normalization 'amp' into 3d density normalization
        rho.

        :param amp:
        :return: rho
        """
        factor = np.pi
        return amp / 2 / factor



import numpy as np
from lenstronomy.Util import image_util


from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
class ImageNoise(object):
    """Class that deals with noise properties of imaging data."""

    def __init__(
        self,
        image_data,
        exposure_time=None,
        background_rms=None,
        noise_map=None,
        gradient_boost_factor=None,
        verbose=True,
        flux_scaling=1,
    ):
        """

        :param image_data: numpy array, pixel data values
        :param exposure_time: int or array of size the data; exposure time
         (common for all pixels or individually for each individual pixel)
        :param background_rms: root-mean-square value of Gaussian background noise
        :param noise_map: int or array of size the data; joint noise sqrt(variance) of each individual pixel.
         Overwrites meaning of background_rms and exposure_time.
        :param gradient_boost_factor: None or float, variance terms added in quadrature scaling with
         gradient^2 * gradient_boost_factor
        :param flux_scaling: scales the model amplitudes to match the imaging data units. This can be used, for example,
         when modeling multiple exposures that have different magnitude zero points (or flux normalizations) but demand
         the same model normalization
        :type flux_scaling: float or int (default=1)
        """
        if exposure_time is not None:
            # make sure no negative exposure values are present no dividing by zero
            if isinstance(exposure_time, int) or isinstance(exposure_time, float):
                if exposure_time <= 10 ** (-10):
                    exposure_time = 10 ** (-10)
            else:
                exposure_time[exposure_time <= 10 ** (-10)] = 10 ** (-10)
        self._exp_map = exposure_time
        self._background_rms = background_rms
        self._noise_map = noise_map
        self.flux_scaling = flux_scaling
        if noise_map is not None:
            assert np.shape(noise_map) == np.shape(image_data)
        else:
            if background_rms is not None and exposure_time is not None:
                if np.any(background_rms * exposure_time) < 1 and verbose is True:
                    print(
                        "WARNING! sigma_b*f %s < 1 count may introduce unstable error estimates with a Gaussian"
                        " error function for a Poisson distribution with mean < 1."
                        % (background_rms * np.max(exposure_time))
                    )
        self._data = image_data
        self._gradient_boost_factor = gradient_boost_factor

    @property
    def background_rms(self):
        """

        :return: rms value of background noise
        """
        if self._background_rms is None:
            if self._noise_map is None:
                raise ValueError(
                    "rms background value as 'background_rms' not specified!"
                )
            self._background_rms = np.median(self._noise_map)
        return self._background_rms

    @property
    def exposure_map(self):
        """
        Units of data and exposure map should result in:
        number of flux counts = data * exposure_map

        :return: exposure map for each pixel
        """
        if self._exp_map is None:
            if self._noise_map is None:
                raise ValueError(
                    "Exposure map has not been specified in Noise() class!"
                )
        return self._exp_map

    @property
    def C_D(self):
        """
        Covariance matrix of all pixel values in 2d numpy array (only diagonal component)
        The covariance matrix is estimated from the data.
        WARNING: For low count statistics, the noise in the data may lead to biased estimates of the covariance matrix.

        :return: covariance matrix of all pixel values in 2d numpy array (only diagonal component).
        """
        if not hasattr(self, "_C_D"):
            if self._noise_map is not None:
                self._C_D = self._noise_map**2
            else:
                self._C_D = covariance_matrix(
                    self._data,
                    self.background_rms,
                    self.exposure_map,
                    self._gradient_boost_factor,
                )
        return self._C_D

    def C_D_model(self, model):
        """

        :param model: model (same as data but without noise)
        :return: estimate of the noise per pixel based on the model flux
        """
        if self._noise_map is not None:
            return self._noise_map**2
        else:
            return covariance_matrix(
                model, self._background_rms, self._exp_map, self._gradient_boost_factor
            )


@export
def covariance_matrix(data, background_rms, exposure_map, gradient_boost_factor=None):
    """Returns a diagonal matrix for the covariance estimation which describes the
    error.

    Notes:

    - the exposure map must be positive definite. Values that deviate too much from the mean exposure time will be
        given a lower limit to not under-predict the Poisson component of the noise.

    - the data must be positive semi-definite for the Poisson noise estimate.
        Values < 0 (Possible after mean subtraction) will not have a Poisson component in their noise estimate.


    :param data: data array, eg in units of photons/second
    :param background_rms: background noise rms, eg. in units (photons/second)^2
    :param exposure_map: exposure time per pixel, e.g. in units of seconds
    :param gradient_boost_factor: None or float, variance terms added in quadrature scaling with
         gradient^2 * gradient_boost_factor
    :return: len(d) x len(d) matrix that give the error of background and Poisson components; (photons/second)^2
    """
    if gradient_boost_factor is not None:
        gradient_map = image_util.gradient_map(data) * gradient_boost_factor
    else:
        gradient_map = 0
    d_pos = np.zeros_like(data)
    d_pos[data >= 0] = data[data >= 0]
    sigma = d_pos / exposure_map + background_rms**2 + gradient_map**2
    return sigma



#!/usr/bin/env python

import os
import sys
from setuptools.command.test import test as TestCommand
from setuptools import find_packages

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


if sys.argv[-1] == "publish":
    os.system("python setup.py sdist upload")
    sys.exit()


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.test_args)
        sys.exit(errno)


readme = open("README.rst").read()
doclink = """
Documentation
-------------

The full documentation can be generated with Sphinx"""

history = open("HISTORY.rst").read().replace(".. :changelog:", "")

desc = open("README.rst").read()
requires = [
    "numpy>=1.13",
    "scipy>=0.19.1",
    "configparser",
    "astropy",
    "mpmath",
    "matplotlib",
    "scikit-learn",
    "numba>=0.43.1",
    "corner>=2.2.1",
    "scikit-image",
    "pyyaml",
    "h5py",
    "pyxdg",
    "schwimmbad",
    "multiprocess>=0.70.8",
]
tests_require = [
    "pytest>=2.3",
    "mock",
    "colossus==1.3.0",
    "slitronomy==0.3.2",
    "emcee>=3.0.0",
    "dynesty",
    "nestcheck",
    "pymultinest",
    "zeus-mcmc>=2.4.0",
    "nautilus-sampler>=0.2.1",
    "coolest",
]

PACKAGE_PATH = os.path.abspath(os.path.join(__file__, os.pardir))


setup(
    name="lenstronomy",
    version="1.11.5",
    description="Strong lens modeling package.",
    long_description=desc,
    author="lenstronomy developers",
    author_email="lenstronomy-dev@googlegroups.com",
    url="https://github.com/lenstronomy/lenstronomy",
    download_url="https://github.com/lenstronomy/lenstronomy/archive/1.11.5.tar.gz",
    packages=find_packages(PACKAGE_PATH, "test"),
    package_dir={"lenstronomy": "lenstronomy"},
    include_package_data=True,
    # setup_requires=requires,
    install_requires=requires,
    license="BSD-3",
    zip_safe=False,
    keywords="lenstronomy",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    tests_require=tests_require,
    cmdclass={"test": PyTest},  # 'build_ext':build_ext,
)



__author__ = "sibirrer"

# this file contains a class to compute the truncated Navaro-Frank-White function (Baltz et al 2009)in mass/kappa space
# the potential therefore is its integral

import numpy as np
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
from lenstronomy.LensModel.Profiles.nfw import NFW

__all__ = ["TNFW"]


class TNFW(LensProfileBase):
    """This class contains functions concerning the truncated NFW profile with a
    truncation function (r_trunc^2)*(r^2+r_trunc^2)

    density equation is:

    .. math::
        \\rho(r) = \\frac{r_\\text{trunc}^2}{r^2+r_\\text{trunc}^2}\\frac{\\rho_0(\\alpha_{R_s})}{r/R_s(1+r/R_s)^2}

    relation are: R_200 = c * Rs
    """

    profile_name = "TNFW"
    param_names = ["Rs", "alpha_Rs", "r_trunc", "center_x", "center_y"]
    lower_limit_default = {
        "Rs": 0,
        "alpha_Rs": 0,
        "r_trunc": 0,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "Rs": 100,
        "alpha_Rs": 10,
        "r_trunc": 100,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        """"""
        self._s = 0.001
        super(LensProfileBase, self).__init__()

    def function(self, x, y, Rs, alpha_Rs, r_trunc, center_x=0, center_y=0):
        """

        :param x: angular position
        :param y: angular position
        :param Rs: angular turn over point
        :param alpha_Rs: deflection at Rs
        :param r_trunc: truncation radius
        :param center_x: center of halo
        :param center_y: center of halo
        :return: lensing potential
        """
        rho0_input = self.alpha2rho0(alpha_Rs=alpha_Rs, Rs=Rs)
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        R = np.maximum(R, self._s * Rs)
        f_ = self.nfwPot(R, Rs, rho0_input, r_trunc)

        return f_

    def _L(self, x, tau):
        """Logarithm that appears frequently.

        :param x: r/Rs
        :param tau: t/Rs
        :return:
        """
        x = np.maximum(x, self._s)
        return np.log(x * (tau + np.sqrt(tau**2 + x**2)) ** -1)

    def F(self, x):
        """Classic NFW function in terms of arctanh and arctan.

        :param x: r/Rs
        :return:
        """

        if isinstance(x, np.ndarray):
            x = np.maximum(x, self._s)
            nfwvals = np.zeros_like(x)
            inds1 = np.where(x < 1)
            inds2 = np.where(x > 1)
            inds3 = np.where(x == 1)
            nfwvals[inds1] = (1 - x[inds1] ** 2) ** -0.5 * np.arctanh(
                (1 - x[inds1] ** 2) ** 0.5
            )
            nfwvals[inds2] = (x[inds2] ** 2 - 1) ** -0.5 * np.arctan(
                (x[inds2] ** 2 - 1) ** 0.5
            )
            nfwvals[inds3] = 1
            return nfwvals

        elif isinstance(x, float) or isinstance(x, int):
            x = np.maximum(x, 0)
            if x == 1:
                return 1
            elif x == 0:
                return 0
            elif x < 1:
                return (1 - x**2) ** -0.5 * np.arctanh((1 - x**2) ** 0.5)
            else:
                return (x**2 - 1) ** -0.5 * np.arctan((x**2 - 1) ** 0.5)

    def derivatives(self, x, y, Rs, alpha_Rs, r_trunc, center_x=0, center_y=0):
        """Returns df/dx and df/dy of the function (integral of TNFW), which are the
        deflection angles.

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param Rs: turn over point in the slope of the NFW profile in angular unit
        :param alpha_Rs: deflection (angular units) at projected Rs
        :param r_trunc: truncation radius (angular units)
        :param center_x: center of halo (in angular units)
        :param center_y: center of halo (in angular units)
        :return: deflection angle in x, deflection angle in y
        """
        rho0_input = self.alpha2rho0(alpha_Rs=alpha_Rs, Rs=Rs)
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        R = np.maximum(R, self._s * Rs)
        f_x, f_y = self.nfwAlpha(R, Rs, rho0_input, r_trunc, x_, y_)
        return f_x, f_y

    def hessian(self, x, y, Rs, alpha_Rs, r_trunc, center_x=0, center_y=0):
        """Returns d^2f/dx^2, d^2f/dxdy, d^2f/dydx, d^2f/dy^2 of the TNFW potential f.

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param Rs: turn over point in the slope of the NFW profile in angular unit
        :param alpha_Rs: deflection (angular units) at projected Rs
        :param r_trunc: truncation radius (angular units)
        :param center_x: center of halo (in angular units)
        :param center_y: center of halo (in angular units)
        :return: Hessian matrix of function d^2f/dx^2, d^f/dy^2, d^2/dxdy
        """

        rho0_input = self.alpha2rho0(alpha_Rs=alpha_Rs, Rs=Rs)
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        R = np.maximum(R, self._s * Rs)

        kappa = self.density_2d(x_, y_, Rs, rho0_input, r_trunc)
        gamma1, gamma2 = self.nfwGamma(R, Rs, rho0_input, r_trunc, x_, y_)
        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        return f_xx, f_xy, f_xy, f_yy

    @staticmethod
    def density(r, Rs, rho0, r_trunc):
        """Three dimensional truncated NFW profile.

        :param r: radius of interest
        :type r: float/numpy array
        :param Rs: scale radius
        :type Rs: float > 0
        :param r_trunc: truncation radius (angular units)
        :type r_trunc: float > 0
        :return: rho(r) density
        """
        return (
            (r_trunc**2 * (r_trunc**2 + r**2) ** -1)
            * rho0
            / (r / Rs * (1 + r / Rs) ** 2)
        )

    def density_2d(self, x, y, Rs, rho0, r_trunc, center_x=0, center_y=0):
        """Projected two dimensional NFW profile (kappa*Sigma_crit)

        :param R: projected radius of interest
        :type R: float/numpy array
        :param Rs: scale radius
        :type Rs: float
        :param rho0: density normalization (characteristic density)
        :type rho0: float
        :param r_trunc: truncation radius (angular units)
        :type r_trunc: float > 0
        :return: Epsilon(R) projected density at radius R
        """
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        x = R * Rs**-1
        tau = float(r_trunc) * Rs**-1
        Fx = self._F(x, tau)
        return 2 * rho0 * Rs * Fx

    def mass_3d(self, r, Rs, rho0, r_trunc):
        """Mass enclosed a 3d sphere or radius r.

        :param r: 3d radius
        :param Rs: scale radius
        :param rho0: density normalization (characteristic density)
        :param r_trunc: truncation radius (angular units)
        :return: M(<r)
        """

        x = r * Rs**-1
        x = np.maximum(x, self._s)
        func = (
            r_trunc**2
            * (
                -2 * x * (1 + r_trunc**2)
                + 4 * (1 + x) * r_trunc * np.arctan(x / r_trunc)
                - 2 * (1 + x) * (-1 + r_trunc**2) * np.log(Rs)
                + 2 * (1 + x) * (-1 + r_trunc**2) * np.log(Rs * (1 + x))
                + 2 * (1 + x) * (-1 + r_trunc**2) * np.log(Rs * r_trunc)
                - (1 + x)
                * (-1 + r_trunc**2)
                * np.log(Rs**2 * (x**2 + r_trunc**2))
            )
        ) / (2.0 * (1 + x) * (1 + r_trunc**2) ** 2)

        m_3d = 4 * np.pi * Rs**3 * rho0 * func
        return m_3d

    def nfwPot(self, R, Rs, rho0, r_trunc):
        """Lensing potential of truncated NFW profile.

        :param R: radius of interest
        :type R: float/numpy array
        :param Rs: scale radius
        :type Rs: float
        :param rho0: density normalization (characteristic density)
        :type rho0: float
        :param r_trunc: truncation radius (angular units)
        :type r_trunc: float > 0
        :return: lensing potential
        """
        x = R / Rs
        x = np.maximum(x, self._s)
        tau = float(r_trunc) / Rs
        hx = self._h(x, tau)
        return 2 * rho0 * Rs**3 * hx

    def nfwAlpha(self, R, Rs, rho0, r_trunc, ax_x, ax_y):
        """Deflection angel of NFW profile along the projection to coordinate axis.

        :param R: radius of interest
        :type R: float/numpy array
        :param Rs: scale radius
        :type Rs: float
        :param rho0: density normalization (characteristic density)
        :type rho0: float
        :param r_trunc: truncation radius (angular units)
        :type r_trunc: float > 0
        :param axis: projection to either x- or y-axis
        :type axis: same as R
        :return:
        """
        R = np.maximum(R, self._s * Rs)
        x = R / Rs
        x = np.maximum(x, self._s)
        tau = float(r_trunc) / Rs
        gx = self._g(x, tau)
        a = 4 * rho0 * Rs * gx / x**2
        return a * ax_x, a * ax_y

    def nfwGamma(self, R, Rs, rho0, r_trunc, ax_x, ax_y):
        """Shear gamma of NFW profile (times Sigma_crit) along the projection to
        coordinate 'axis'.

        :param R: radius of interest
        :type R: float/numpy array
        :param Rs: scale radius
        :type Rs: float
        :param rho0: density normalization (characteristic density)
        :type rho0: float
        :param r_trunc: truncation radius (angular units)
        :type r_trunc: float > 0
        :param axis: projection to either x- or y-axis
        :type axis: same as R
        :return:
        """
        R = np.maximum(R, self._s * Rs)
        x = R / Rs
        tau = float(r_trunc) * Rs**-1
        gx = self._g(x, tau)
        Fx = self._F(x, tau)
        a = 2 * rho0 * Rs * (2 * gx / x**2 - Fx)
        return a * (ax_y**2 - ax_x**2) / R**2, -a * 2 * (ax_x * ax_y) / R**2

    def mass_2d(self, R, Rs, rho0, r_trunc):
        """Analytic solution of the projection integral (convergence)

        :param R: projected radius
        :param Rs: scale radius
        :param rho0: density normalization (characteristic density)
        :param r_trunc: truncation radius (angular units)
        :return: mass enclosed 2d projected cylinder
        """

        x = R / Rs
        x = np.maximum(x, self._s)
        tau = r_trunc / Rs
        gx = self._g(x, tau)
        m_2d = 4 * rho0 * Rs * R**2 * gx / x**2 * np.pi
        return m_2d

    def _F(self, X, tau):
        """Analytic solution of the projection integral (convergence)

        :param X: R/Rs
        :type X: float >0
        """
        t2 = tau**2
        X = np.maximum(X, self._s)
        _F = self.F(X)
        a = t2 * (t2 + 1) ** -2
        if isinstance(X, np.ndarray):
            # b = (t2 + 1) * (X ** 2 - 1) ** -1 * (1 - _F)
            b = np.ones_like(X)
            b[X == 1] = (t2 + 1) * 1.0 / 3
            b[X != 1] = (t2 + 1) * (X[X != 1] ** 2 - 1) ** -1 * (1 - _F[X != 1])

        elif isinstance(X, float) or isinstance(X, int):
            if X == 1:
                b = (t2 + 1) * 1.0 / 3
            else:
                b = (t2 + 1) * (X**2 - 1) ** -1 * (1 - _F)
        else:
            raise ValueError(
                "The variable type is not compatible with the function, please use float,"
                " int or ndarray's."
            )

        c = 2 * _F
        d = -np.pi * (t2 + X**2) ** -0.5
        e = (t2 - 1) * (tau * (t2 + X**2) ** 0.5) ** -1 * self._L(X, tau)
        result = a * (b + c + d + e)

        return result

    def _g(self, x, tau):
        """Analytic solution of integral for NFW profile to compute deflection angel and
        gamma.

        :param x: R/Rs
        :type x: float >0
        """
        x = np.maximum(x, self._s)
        return (
            tau**2
            * (tau**2 + 1) ** -2
            * (
                (tau**2 + 1 + 2 * (x**2 - 1)) * self.F(x)
                + tau * np.pi
                + (tau**2 - 1) * np.log(tau)
                + np.sqrt(tau**2 + x**2)
                * (-np.pi + self._L(x, tau) * (tau**2 - 1) * tau**-1)
            )
        )

    @staticmethod
    def _cos_function(x):
        if isinstance(x, np.ndarray) or isinstance(x, list):
            out = np.empty_like(x)
            inds1 = np.where(x < 1)
            inds2 = np.where(x >= 1)

            out[inds1] = -np.arccosh(1 / x[inds1]) ** 2
            out[inds2] = np.arccos(1 / x[inds2]) ** 2

        elif isinstance(x, float) or isinstance(x, int):
            if x < 1:
                out = -np.arccosh(1 / x) ** 2
            else:
                out = np.arccos(1 / x) ** 2

        else:
            raise Exception("x data type %s not recognized." % x)

        return out

    def _h(self, x, tau):
        """Expression for the integral to compute potential.

        :param x: R/Rs
        :param tau: r_trunc/Rs
        :type x: float >0
        """
        x = np.maximum(x, self._s)

        u = x**2
        t2 = tau**2
        Lx = self._L(x, tau)
        Fx = self.F(x)

        return (t2 + 1) ** -2 * (
            2
            * t2
            * np.pi
            * (tau - (t2 + u) ** 0.5 + tau * np.log(tau + (t2 + u) ** 0.5))
            + 2 * (t2 - 1) * tau * (t2 + u) ** 0.5 * Lx
            + t2 * (t2 - 1) * Lx**2
            + 4 * t2 * (u - 1) * Fx
            + t2 * (t2 - 1) * self._cos_function(x)
            + t2 * ((t2 - 1) * np.log(tau) - t2 - 1) * np.log(u)
            - t2
            * (
                (t2 - 1) * np.log(tau) * np.log(4 * tau)
                + 2 * np.log(0.5 * tau)
                - 2 * tau * (tau - np.pi) * np.log(tau * 2)
            )
        )

    @staticmethod
    def alpha2rho0(alpha_Rs, Rs):
        """Convert angle at Rs into rho0; neglects the truncation.

        :param alpha_Rs: deflection angle at RS
        :param Rs: scale radius
        :return: density normalization (characteristic density)
        """
        return NFW.alpha2rho0(alpha_Rs, Rs)

    @staticmethod
    def rho02alpha(rho0, Rs):
        """Convert rho0 to angle at Rs; neglects the truncation.

        :param rho0: density normalization (characteristic density)
        :param Rs: scale radius
        :return: deflection angle at RS
        """
        return NFW.rho02alpha(rho0, Rs)



import pytest
import numpy as np
import numpy.testing as npt
from lenstronomy.Util import util
from lenstronomy.LightModel.Profiles.gaussian import (
    MultiGaussian,
    MultiGaussianEllipse,
    GaussianEllipse,
    Gaussian,
)


class TestMultiGaussian(object):
    """Class to test the Gaussian profile."""

    def setup_method(self):
        pass

    def test_function_split(self):
        """

        :return:
        """
        profile = MultiGaussian()
        output = profile.function_split(
            x=1.0, y=1.0, amp=[1.0, 2], sigma=[1, 2], center_x=0, center_y=0
        )
        npt.assert_almost_equal(output[0], 0.058549831524319168, decimal=8)
        npt.assert_almost_equal(output[1], 0.061974997154826489, decimal=8)


class TestGaussian(object):
    def setup_method(self):
        pass

    def test_total_flux(self):
        gauss = Gaussian()
        deltapix = 0.1
        amp = 1
        x_grid, y_gird = util.make_grid(100, deltapix=deltapix)
        flux = gauss.function(x_grid, y_gird, amp=amp, sigma=1)
        flux_integral = np.sum(flux) * deltapix**2
        npt.assert_almost_equal(flux_integral, amp, decimal=3)
        # make grid
        # sum grid
        # evaluate total flux


class TestGaussianEllipse(object):
    def setup_method(self):
        pass

    def test_function_split(self):
        """

        :return:
        """
        multiGaussian = MultiGaussian()
        multiGaussianEllipse = MultiGaussianEllipse()
        output = multiGaussian.function_split(
            x=1.0, y=1.0, amp=[1.0, 2], sigma=[1, 2], center_x=0, center_y=0
        )
        output_2 = multiGaussianEllipse.function_split(
            x=1.0, y=1.0, amp=[1.0, 2], sigma=[1, 2], e1=0, e2=0, center_x=0, center_y=0
        )
        npt.assert_almost_equal(output[0], output_2[0], decimal=8)
        npt.assert_almost_equal(output[1], output_2[1], decimal=8)

    def test_gaussian_ellipse(self):
        gaussianEllipse = GaussianEllipse()
        gaussian = Gaussian()
        sigma = 1
        flux = gaussianEllipse.function(1, 1, amp=1, sigma=sigma, e1=0, e2=0)
        flux_spherical = gaussian.function(1, 1, amp=1, sigma=sigma)
        npt.assert_almost_equal(flux, flux_spherical, decimal=8)

    def test_light_3d(self):
        gaussianEllipse = GaussianEllipse()
        gaussian = Gaussian()

        sigma = 1
        r = 1.0
        amp = 1.0
        flux_spherical = gaussian.light_3d(r, amp, sigma)
        flux = gaussianEllipse.light_3d(r, amp, sigma)
        npt.assert_almost_equal(flux, flux_spherical, decimal=8)

        multiGaussian = MultiGaussian()
        multiGaussianEllipse = MultiGaussianEllipse()
        amp = [1, 2]
        sigma = [1.0, 2]
        flux_spherical = multiGaussian.light_3d(r, amp, sigma)
        flux = multiGaussianEllipse.light_3d(r, amp, sigma)
        npt.assert_almost_equal(flux, flux_spherical, decimal=8)


if __name__ == "__main__":
    pytest.main()



import types


def test_short_and_laconic():
    import lenstronomy as ls

    ls.short()

    # We can access submodules as symbols...
    assert hasattr(ls, "LensModel")
    assert isinstance(ls.LensModel, types.ModuleType)

    # ... also recursively.
    assert hasattr(ls.LensModel, "lens_model")
    assert isinstance(ls.LensModel.lens_model, types.ModuleType)

    # Symbols inside the module are available
    assert "LensModel" in ls.LensModel.lens_model.__all__
    assert hasattr(ls.LensModel.lens_model, "LensModel")
    assert isinstance(ls.LensModel.lens_model.LensModel, type)

    ls.laconic()

    # We can access lenstronomy symbols directy
    assert hasattr(ls, "MultiBandImageReconstruction")
    assert isinstance(ls.MultiBandImageReconstruction, type)

    # Non-lenstronomy symbols are not accessible
    assert not hasattr(ls, "np")
    assert not hasattr(ls, "numpy")

    # Clashing symbols are not accessible
    assert not hasattr(ls, "PSF")

    # 'short' type access still works
    assert hasattr(ls, "Analysis")
    assert isinstance(ls.Analysis, types.ModuleType)
    assert hasattr(ls.Analysis, "image_reconstruction")
    assert isinstance(ls.Analysis.image_reconstruction, types.ModuleType)
    assert hasattr(ls.Analysis.image_reconstruction, "MultiBandImageReconstruction")
    assert isinstance(
        ls.Analysis.image_reconstruction.MultiBandImageReconstruction, type
    )

    # Key classes clashing with submodule names are
    # accessible with an extra underscore
    assert hasattr(ls, "LensModel")
    assert isinstance(ls.LensModel, types.ModuleType)
    assert hasattr(ls, "LensModel_")
    assert isinstance(ls.LensModel_, type)



import unittest
import pytest
from lenstronomy.LightModel.Profiles.profile_base import LightProfileBase


class TestRaise(unittest.TestCase):
    def test_raise(self):
        lighModel = LightProfileBase()
        with self.assertRaises(ValueError):
            lighModel.function()
        with self.assertRaises(ValueError):
            lighModel.light_3d()


if __name__ == "__main__":
    pytest.main()



from lenstronomy.PointSource.Types.base_ps import PSBase
import numpy as np

__all__ = ["Unlensed"]


class Unlensed(PSBase):
    """
    class of a single point source in the image plane, aka star
    Name within the PointSource module: 'UNLENSED'
    This model can deal with arrays of point sources.
    parameters: ra_image, dec_image, point_amp

    """

    def image_position(self, kwargs_ps, **kwargs):
        """On-sky position.

        :param kwargs_ps: keyword argument of point source model
        :return: numpy array of x, y image positions
        """
        ra_image = kwargs_ps["ra_image"]
        dec_image = kwargs_ps["dec_image"]
        return np.array(ra_image), np.array(dec_image)

    def source_position(self, kwargs_ps, **kwargs):
        """Original physical position (identical for this object)

        :param kwargs_ps: keyword argument of point source model
        :return: numpy array of x, y source positions
        """
        ra_image = kwargs_ps["ra_image"]
        dec_image = kwargs_ps["dec_image"]
        return np.array(ra_image), np.array(dec_image)

    def image_amplitude(self, kwargs_ps, **kwargs):
        """Amplitudes as observed on the sky.

        :param kwargs_ps: keyword argument of point source model
        :param kwargs: keyword arguments of function call (which are not used for this
            object
        :return: numpy array of amplitudes
        """
        point_amp = kwargs_ps["point_amp"]
        return np.array(point_amp)

    def source_amplitude(self, kwargs_ps, **kwargs):
        """Intrinsic source amplitudes.

        :param kwargs_ps: keyword argument of point source model
        :param kwargs: keyword arguments of function call (which are not used for this
            object
        :return: numpy array of amplitudes
        """
        point_amp = kwargs_ps["point_amp"]
        return np.array(point_amp)
